import os
from dotenv import load_dotenv
from .RDBMSTypes import JobServers
from .Database import SQLAlchemyConnection


class Configuration(object):
    """
        Handle Node Configuration
    """

    _config = {}
    if os.name == 'nt':
        grease_dir = "C:\\grease"
    else:
        grease_dir = "/var/tmp/grease"
    fs_Separator = os.sep
    op_name = os.name
    grease_log = grease_dir + os.sep + "grease.log"
    identity_file = grease_dir + os.sep + "grease_identity.txt"
    identity = None
    _node_db_id = None

    def __init__(self):
        # Ensure the GREASE Dir
        if not os.path.isdir(self.grease_dir):
            os.mkdir(self.grease_dir)
        # load up config
        self._load_config()

    @staticmethod
    def generate():
        # type: () -> Configuration
        return Configuration()

    @staticmethod
    def node_identity():
        # type: () -> str
        if os.path.isfile(Configuration.identity_file):
            fil = open(Configuration.identity_file, "r")
            identity = fil.read().rstrip()
            fil.close()
        else:
            identity = ""
        return identity

    def node_db_id(self):
        # type: () -> int
        if not self._node_db_id:
            identity = Configuration.node_identity()
            conn = SQLAlchemyConnection(Configuration())
            result = conn.get_session().query(JobServers).filter(JobServers.host_name == identity).first()
            self._node_db_id = result.id
        return self._node_db_id

    def get(self, key, default=None):
        # type: (str, str) -> object
        return self._config.get(key, default)

    def _load_config(self):
        # type: () -> None
        # load optional config file
        if os.path.isfile(self.grease_dir + os.sep + "grease.conf"):
            load_dotenv(self.grease_dir + os.sep + "grease.conf", override=True)
        # Load default Environment
        self._config = os.environ
        # Load Identity
        self.identity = self.node_identity()



from collections import defaultdict
from tgt_grease_core_util.RDBMSTypes import JobServers, SourceData
from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand
from tgt_grease_core_util.ImportTools import Importer
from tgt_grease_core_util import SQLAlchemyConnection, Configuration
from . import Detectors
from .Detectors import DetectorConfiguration
from sqlalchemy import update
from datetime import datetime


class Detector(GreaseDaemonCommand):
    def __init__(self):
        super(Detector, self).__init__()
        self._config = Configuration()
        self._sql = SQLAlchemyConnection(self._config)
        self._excelsior_config = DetectorConfiguration.ConfigurationLoader()
        self._importer = Importer(self._ioc.message())

    def execute(self, context='{}'):
        # first lets see if we have some stuff to parse
        result = self._sql.get_session().query(SourceData, JobServers)\
            .filter(JobServers.id == SourceData.detection_server)\
            .filter(JobServers.id == self._config.node_db_id())\
            .filter(SourceData.detection_start_time == None)\
            .filter(SourceData.detection_end_time == None)\
            .filter(SourceData.detection_complete == False)\
            .limit(15)\
            .all()
        if not result:
            self._ioc.message().debug("No sources scheduled for detection on this node", True)
            return True
        else:
            # Now lets loop through
            self._ioc.message().debug("TOTAL SOURCE DOCUMENTS RETURNED: [{0}]".format(len(result)), True)
            for source in result:
                # first claiming them as ours then parsing them
                self._take_ownership(source.SourceData.id)
                # now lets parse the sources
                self._ioc.message().debug("PROCESSING SOURCE ID: [{0}]".format(source.SourceData.id), True)
                parsed_source = self._parse_source(
                    source.SourceData.source_data,
                    self._excelsior_config.get_scanner_config(source.SourceData.scanner)
                )
                self._complete(source.SourceData.id)
                # now lets assign this parsed source out
                if self._schedule_scheduling(source.SourceData.id, parsed_source):
                    self._ioc.message().info("Successfully Scheduled Parsed Source ID: [" + str(source.SourceData.id) + "]")
                    continue
                else:
                    self._reverse(source.SourceData.id)
                    self._ioc.message().error("Failed To Schedule Parsed Source ID: [" + str(source.SourceData.id) + "]")
                    continue
        return True

    def _take_ownership(self, source_file_id):
        # type: (int) -> None
        stmt = update(SourceData)\
            .where(SourceData.id == source_file_id)\
            .values(detection_start_time=datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug("TAKING OWNERSHIP OF SOURCE ID: [{0}]".format(source_file_id), True)

    def _complete(self, source_file_id):
        # type: (int) -> None
        stmt = update(SourceData)\
            .where(SourceData.id == source_file_id)\
            .values(detection_complete=True, detection_end_time=datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug("COMPLETING SOURCE ID: [{0}]".format(source_file_id), True)

    def _reverse(self, source_file_id):
        stmt = update(SourceData)\
            .where(SourceData.id == source_file_id)\
            .values(detection_start_time=None, detection_end_time=None, detection_complete=None)
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug("SOURCE FILE FAILED SCHEDULING REVERSING FILE: [{0}]".format(source_file_id), True)

    def _parse_source(self, sources, rule_set):
        # type: (dict, list) -> dict
        # lets parse the source
        # now lets get the data out of it
        # now lets loop through those results
        # first lets setup some state tracking stuff
        final_source_data = defaultdict()
        index = 0
        total_records = len(sources)
        for source in sources:
            self._ioc.message().debug(
                "PROCESSING OBJECT [{0}] OF [{1}] INTERNAL STRUCTURE: [{2}]".format(
                    str(index + 1),
                    str(total_records),
                    str(source)
                ),
                True
            )
            # first lets initialize all the source data into the object
            final_source_data[index] = source
            # now lets create our rule_processing key
            final_source_data[index]['rule_processing'] = defaultdict()
            # now lets loop through the rules we want to parse this source with
            for rule in rule_set:
                self._ioc.message().debug("PROCESSING RULE [{0}]".format(str(rule['name'])), True)
                # first lets make the rule processing result key
                final_source_data[index]['rule_processing'][rule['name']] = defaultdict()
                # next lets compute each rule in the rule processor
                # Yes another for loop but I promise this will be fast
                # look at some point we have to iterate over data
                for detector, detector_config in rule['logic'].iteritems():
                    # lets try to new up this detector
                    detector_instance = self._importer.load('tgt_grease_enterprise.Detectors', detector, True)
                    if isinstance(detector_instance, Detectors.BaseDetector):
                        self._ioc.message().debug(
                            "PROCESSING RULE [{0}] LOGICAL BLOCK [{1}]".format(str(rule['name']), str(detector)),
                            True
                        )
                        # we have a valid detector to parse with
                        # lets compute the source with the detector config
                        detector_instance.param_compute(source, detector_config)
                        # lets observe the result
                        result = detector_instance.get_result()
                        if result['result']:
                            # we passed the rule lets set that in the final source data
                            self._ioc.message().debug(
                                "PROCESSING RULE [{0}] LOGICAL BLOCK [{1}] PASSED".format(
                                    str(rule['name']),
                                    str(detector)
                                ),
                                True
                            )
                            final_source_data[index]['rule_processing'][rule['name']]['status'] = True
                            self._ioc.message().debug("RULE [{0}] PASSED THUS FAR".format(str(rule['name'])), True)
                            result.pop('result')
                            if 'parameters' not in final_source_data[index]['rule_processing'][rule['name']]:
                                final_source_data[index]['rule_processing'][rule['name']]['parameters'] = result
                            else:
                                final_source_data[index]['rule_processing'][rule['name']]['parameters'].update(result)
                            if 'constants' in rule:
                                final_source_data[index]['rule_processing'][rule['name']]['parameters']['constants'] = rule['constants']
                            # del out the instance
                            del detector_instance
                            # route on
                            continue
                        else:
                            # rule failed
                            final_source_data[index]['rule_processing'][rule['name']]['status'] = False
                            self._ioc.message().debug(
                                "PROCESSING RULE [{0}] LOGICAL BLOCK [{1}] FAILED :: SOURCE FAILS RULE SET".format(
                                    str(rule['name']),
                                    str(detector)
                                ),
                                True
                            )
                            # del out the instance
                            del detector_instance
                            # route on
                            break
                    else:
                        # we got an invalid detector and it couldn't be found
                        self._ioc.message().error("Invalid Detector: [" + str(detector) + "]", hipchat=True)
                        del detector_instance
                        break
                # finally lets convert back to normal dict for the rule
                final_source_data[index]['rule_processing'][rule['name']] = dict(
                    final_source_data[index]['rule_processing'][rule['name']]
                )
            # now lets convert the rule_processing back to a normal array
            final_source_data[index]['rule_processing'] = dict(final_source_data[index]['rule_processing'])
            self._ioc.message().debug(
                "FINAL SOURCE RULE PROCESSING STRUCTURE: [{0}]".format(
                    str(final_source_data[index]['rule_processing'])
                ),
                True
            )
            self._ioc.message().debug(
                "PROCESSING OBJECT [{0}] OF [{1}] COMPLETE".format(
                    str(index + 1),
                    str(total_records)
                ),
                True
            )
            # finally increment our pointer
            index += 1
        # return final for usage elsewhere
        return final_source_data

    def _schedule_scheduling(self, source_id, updated_source):
        # type: (dict, str)  -> bool
        # first lets get applicable servers to run detectors
        # lets only get the least assigned server so we can round robin
        result = self._sql.get_session()\
            .query(JobServers)\
            .filter(JobServers.scheduler == True)\
            .filter(JobServers.active == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if not result:
            self._ioc.message().error("Failed to find active scheduling server!::Dropping Scan", hipchat=True)
            return False
        else:
            server = result.id
            # Now lets update the sources for the determined server to work
            stmt = update(SourceData)\
                .where(SourceData.id == source_id)\
                .values(scheduling_server=server, source_data=updated_source)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            # finally lets ensure we account for the fact our server is going to do
            # that job and increment the assignment counter
            stmt = update(JobServers).where(JobServers.id == server).values(jobs_assigned=result.jobs_assigned + 1)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            self._ioc.message().debug(
                "DETECTION FOR SOURCE COMPLETE::SCHEDULED TO SERVER [{0}]".format(server),
                True
            )
            return True



from .Scanner import ScanOnConfig
from .Detector import Detector
from .Scheduler import Scheduler
from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand


class CaptainKirk(GreaseDaemonCommand):
    def __init__(self):
        super(CaptainKirk, self).__init__()

    def execute(self, context='{}'):
        instance = ScanOnConfig()
        self._ioc.message().debug("Scanning Sources")
        instance.execute()
        self._ioc.message().debug("Parsing Sources")
        instance = Detector()
        instance.execute()
        self._ioc.message().debug("Scheduling Jobs")
        instance = Scheduler()
        instance.execute()
        return True



from unittest import TestCase
from tgt_grease_core import GreaseRouter
from tgt_grease_core import GreaseCommand


class test_command(GreaseCommand):
    def execute(self):
        return True


class TestRouter(TestCase):
    """class TestRouter tests router logic"""
    def test_is_router(self):
        """test_is_router tests type returned from newing up"""
        router = GreaseRouter.Router()
        self.assertTrue(router, GreaseRouter.Router)



from abc import ABCMeta, abstractmethod


class BaseSource(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def parse_source(self, rule_config):
        # type (dict) -> None
        pass

    @abstractmethod
    def get_records(self):
        # type: () -> dict
        pass

    def duplicate_check_fields(self):
        # type: () -> list
        return []




from unittest import TestCase
from tgt_grease_core_util import ImportTools
import collections
from tgt_grease_core_util import Logging

class test_importTools(TestCase):
    def test_load(self):
        obj = ImportTools.Importer(Logging.Logger())
        loaded_module = obj.load('collections', 'defaultdict', True)
        self.assertEqual(type(loaded_module), collections.defaultdict)
        del obj, loaded_module

    def test_attribute_err(self):
        obj = ImportTools.Importer(Logging.Logger())

        def load_the_mod(obj2):
            obj2.load('collections', 'defeaultdict1', True)
        self.assertRaises(AttributeError, load_the_mod(obj))
        del obj

    def test_import_err(self):
        obj = ImportTools.Importer(Logging.Logger())

        def load_the_mod(obj2):
            obj2.load('collections1', 'defeaultdict', True)
        self.assertRaises(ImportError, load_the_mod(obj))
        del obj



import datetime
from abc import ABCMeta, abstractmethod
from .Database import SQLAlchemyConnection
from .RDBMSTypes import JobTelemetry, JobTelemetryDaemon
from .Configuration import Configuration
import os


class Telemetry:
    __metaclass__ = ABCMeta

    _config = Configuration()
    _sql = SQLAlchemyConnection(_config)

    def __init__(self, command_obj, is_daemon=False):
        # type: (tgt_grease_core.GreaseCommand, bool) -> None
        self._command = command_obj
        self._is_daemon = is_daemon

    @abstractmethod
    def store_result(self, success_state):
        # type: (bool) -> None
        pass


class Database(Telemetry):
    def __init__(self, command_obj, is_daemon):
        # type: (tgt_grease_core.GreaseCommand, bool) -> None
        Telemetry.__init__(self, command_obj, is_daemon)

    def store_result(self, success_state):
        # type: () -> None
        telemetry = JobTelemetry(
            command=self._command.get_exe_state().get('command_id', 1),
            success=str(success_state),
            affected=str(self._command.get_effected()),
            start_time=str(datetime.datetime.fromtimestamp(self._command.get_start_time()).strftime('%Y-%m-%d %H:%M:%S')),
            entry_time=datetime.datetime.utcnow(),
            server_id=self._config.node_db_id()
        )
        self._sql.get_session().add(telemetry)
        self._sql.get_session().commit()


class DatabaseDaemon(Telemetry):
    def __init__(self, Command_Obj):
        # type: (tgt_grease_core.GreaseCommand) -> None
        super(DatabaseDaemon, self).__init__(Command_Obj, Command_Obj.get_exe_state()['isDaemon'])

    def store_result(self, success_state=None):
        telemetry = JobTelemetryDaemon(
            command=self._command.get_exe_state()['command_id'],
            affected=self._command.get_effected(),
            start_time=str(datetime.datetime.fromtimestamp(self._command.get_start_time()).strftime('%Y-%m-%d %H:%M:%S')),
            execution_success=self._command.get_exe_state()['execution'],
            command_success=self._command.get_exe_state()['result'],
            entry_time=datetime.datetime.utcnow(),
            server_id=self._config.node_db_id()
        )
        self._sql.get_session().add(telemetry)
        self._sql.get_session().commit()



from .BaseDetectorClass import BaseDetector
import re

# Basic Regex Detector Config
# ...
# "regex": [
#    {
#       "field": "my_field", <- Mandatory
#       "pattern": 10, <- Mandatory
#       "variable": True, <- Optional Assignment Parameter
#       "variable_name": "my_field" <- Optional Assignment Parameter
#    }
# ]
#
# ...


class regex(BaseDetector):
    def __init__(self):
        self._result = {}

    def param_compute(self, source, regex_conf):
        # type: (dict, dict) -> None
        rules_needed_to_pass = len(regex_conf)
        rules_passed = 0
        final_result = {'result': False}
        for regex_param in regex_conf:
            # loop through the parameter set
            if regex_param['field'] in source:
                # Incase field is None type (Null from JSON) Transform to empty string for operations
                if source[regex_param['field']] is None:
                    source[regex_param['field']] = ''
                # if the field exists in the source record lets work with its
                result = re.findall(regex_param['pattern'], source[regex_param['field']])
                if len(result) > 0:
                    if 'variable' in regex_param:
                        if regex_param['variable'] and 'variable_name' in regex_param:
                            final_result[regex_param['variable_name']] = result
                    rules_passed += 1
                continue
            else:
                # We have no source field to parse
                self._result = {'result': False}
                return None
        if rules_passed == rules_needed_to_pass:
            final_result['result'] = True
            self._result = final_result
            return None
        else:
            self._result = {'result': False}
            return None

    def get_result(self):
        # type: () -> dict
        return self._result



import os
from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand
from .Detectors import DetectorConfiguration
from tgt_grease_core_util import SQLAlchemyConnection, Configuration
from tgt_grease_core_util.ImportTools import Importer
from tgt_grease_core_util.RDBMSTypes import JobServers, SourceData
from tgt_grease_enterprise import BaseSource
from .DeDuplification import SourceDeDuplify
from datetime import datetime
from sqlalchemy import update


class ScanOnConfig(GreaseDaemonCommand):
    def __init__(self):
        super(ScanOnConfig, self).__init__()
        self._config = Configuration()
        self._sql = SQLAlchemyConnection(self._config)
        self._scanner_config = DetectorConfiguration.ConfigurationLoader()
        self._context = '{}'
        self._source_data = {}
        self._importer = Importer(self._ioc.message())
        self._duplification_filter = SourceDeDuplify(self._ioc.message())

    def get_source_data(self):
        # type: () -> dict
        return self._source_data

    def set_source_data(self, data):
        # type: (dict) -> None
        self._source_data = data

    def execute(self, context='{}'):
        # engage scanning
        self.scan()
        # clear up this 
        del self._duplification_filter
        return True

    def scan(self):
        # engage scanners scotty
        for scanner in self._scanner_config.get_scanners():
            self._ioc.message().debug(
                "STARTING SCANNER [{0}]".format(str(scanner)),
                True
            )
            # Ensure if we are focused only to process our source
            if os.getenv('GREASE_SOURCE_FOCUS'):
                # we have one, ensure its on the list of configured scanners
                if str(os.getenv('GREASE_SOURCE_FOCUS')) != str(scanner):
                    # It does not match, continue
                    self._ioc.message().info(
                        "Scanner skipped because not focus: ["
                        + str(scanner)
                        + "] searching for ["
                        + str(os.getenv('GREASE_SOURCE_FOCUS'))
                        + "]"
                    )
                    continue
            # lets loop through our scanners/detectors to execute the parsing
            # try to load the scanner we want
            parser = self._importer.load(os.getenv('GREASE_SOURCES_PKG', ''), scanner, True)
            # type check that bad boy to ensure sanity
            if isinstance(parser, BaseSource):
                # if we got back a valid source lets parse that sucker
                self._ioc.message().debug(
                    "PARSING SOURCE [{0}]".format(str(scanner)),
                    True
                )
                parser.parse_source(self._scanner_config.get_scanner_config(scanner))
                # lets get the results of the parse
                # here we run our de-duplication logic
                self._ioc.message().debug(
                    "PASSING RESULT TO DEDUPLICATION ENGINE [{0}]".format(str(scanner)),
                    True
                )
                source = self._duplification_filter.create_unique_source(
                    scanner,
                    parser.duplicate_check_fields(),
                    list(parser.get_records())
                )
                self._ioc.message().debug(
                    "ATTEMPTING DETECTION SCHEDULING [{0}]".format(str(scanner)),
                    True
                )
                if self._schedule_detection(source, scanner):
                    self._ioc.message().info("Detector job scheduled from scanner: [" + str(scanner) + "]")
                else:
                    self._ioc.message().error("Failed to schedule source detection for [" + str(scanner) + "]", hipchat=True)
                del parser
            else:
                # else something went haywire pls feel free to fix your config
                self._ioc.message().error("Invalid Scanner In Configurations: [" + str(scanner) + "]", hipchat=True)

    def _schedule_detection(self, sources, scanner):
        # type: (dict, str)  -> bool
        # first lets get applicable servers to run detectors
        # lets only get the least assigned server so we can round robin
        result = self._sql.get_session()\
            .query(JobServers)\
            .filter(JobServers.detector == True)\
            .filter(JobServers.active == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if not result:
            self._ioc.message().error("Failed to find detection server! dropping scan!", hipchat=True)
            return False
        else:
            server = result.id
        # Now lets insert the sources for the determined server to work
        source = SourceData(
            source_data=sources,
            source_server=self._config.node_db_id(),
            detection_server=server,
            scanner=scanner,
            created_time=datetime.utcnow()
        )
        self._sql.get_session().add(source)
        self._sql.get_session().commit()
        # finally lets ensure we account for the fact our server is going to do
        # that job and increment the assignment counter
        stmt = update(JobServers).where(JobServers.id == server).values(jobs_assigned=result.jobs_assigned + 1)
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug(
            "SOURCE SCHEDULED FOR DETECTION [{0}] TO SERVER [{1}]".format(str(scanner), str(server)),
            True
        )
        return True



from unittest import TestCase
from tgt_grease_enterprise.Detectors import TimeDelta
from datetime import datetime, timedelta


class TestTimeDelta(TestCase):
    def _setup(self, delta, source_delta, direction='past'):
        self.TimeDelta = TimeDelta()

        if direction is 'past':
            dt = datetime.now() - timedelta(days=source_delta)
        else:
            dt = datetime.now() + timedelta(days=source_delta)

        date_string = dt.strftime("%Y-%m-%d")

        self.source_data = {
            "field": date_string
        }

        self.rules = [
            {
                "field": "field",
                "date_format": "%Y-%m-%d",
                "delta": delta,
                "unit": "days",
                "direction": direction,
                "operator": ">="
            }
        ]

    def _teardown(self):
        if self.TimeDelta:
            del self.TimeDelta
        if self.rules:
            del self.rules
        if self.source_data:
            del self.source_data

    def test_past_true(self):
        self._setup(delta=1, source_delta=3, direction='past')

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertTrue(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_past_false(self):
        self._setup(delta=3, source_delta=1, direction='past')

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_future_true(self):
        self._setup(delta=1, source_delta=3, direction='future')

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertTrue(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_future_false(self):
        self._setup(delta=3, source_delta=1, direction='future')

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_malformed_rule(self):
        self._setup(delta=1, source_delta=3)
        self.rules += {
            'biscuits': 'gravy'
        }

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_bad_unit(self):
        self._setup(delta=1, source_delta=3)
        self.rules[0]['unit'] = 'eons'

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_bad_operator(self):
        self._setup(delta=1, source_delta=3)
        self.rules[0]['operator'] = 'note'

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_date_supplied(self):
        self._setup(delta=3, source_delta=3)
        self.source_data["field"] = "2017-08-20"
        self.rules[0]['date'] = "2017-08-24"

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertTrue(self.TimeDelta.get_result()['result'])
        self._teardown()

    def test_format_mismatch(self):
        self._setup(delta=1, source_delta=3)
        self.rules[0]['date_format'] = '%y%m%d'

        self.TimeDelta.param_compute(source=self.source_data, rules=self.rules)
        self.assertFalse(self.TimeDelta.get_result()['result'])
        self._teardown()



import importlib
from tgt_grease_core_util.Logging import Logger


class Importer(object):
    def __init__(self, logger):
        # type: (Logger) -> None
        self.purpose = "to load modules"
        self._log = logger

    def load(self, module_name, module_class, fully_qualified=False):
        import inspect
        # type: (str, str, bool) -> object
        try:
            if bool(fully_qualified):
                loaded_module = importlib.import_module(str(module_name))
            else:
                loaded_module = importlib.import_module('tgt_grease_' + str(module_name))
            try:
                req = getattr(loaded_module, module_class)
                instance = req()
                return instance
            except AttributeError as a:
                self._log.error("Failed to create instance of class::[" + str(a) + "]")
                return None
        except ImportError as i:
            self._log.error("Failed to Import Module::[" + str(i) + "] [" +
                            str(inspect.getouterframes(inspect.currentframe())[1])
                        + "]")
            return None



#! /usr/bin/env python
import os
import win32serviceutil
import win32service
import win32event
import servicemanager
import socket
import sys
from tgt_grease_daemon import DaemonRouter, UnixDaemon


class AppServerSvc (win32serviceutil.ServiceFramework):

    _svc_name_ = "GreaseDaemon"
    _svc_display_name_ = "GREASE Daemon Server"
    _svc_description_ = "GREASE Async Daemon Server for Automation Work"

    def __init__(self,args):
        win32serviceutil.ServiceFramework.__init__(self,args)
        self.hWaitStop = win32event.CreateEvent(None,0,0,None)
        socket.setdefaulttimeout(60)

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_,''))
        self.main()

    def main(self):
        if 'install' not in sys.argv:
            s = DaemonRouter()
            s.set_process(self)
            s.main()
        else:
            return

    def start(self):
        self.SvcDoRun()

    def stop(self):
        self.SvcStop()

    def restart(self):
        self.SvcStop()
        self.SvcDoRun()

if os.name == 'nt':
        print sys.argv
        if 'install' not in sys.argv:
            win32serviceutil.HandleCommandLine(AppServerSvc, argv=['', 'start'])
        else:
            win32serviceutil.HandleCommandLine(AppServerSvc)
else:
    s = DaemonRouter()
    s.set_process(UnixDaemon(s))
    s.entry_point()



from unittest import TestCase
from tgt_grease_core_util import Configuration
from tgt_grease_core_util import Logging
from logging import Logger


class test_logging(TestCase):

    _config = Configuration()

    def test_is_object_LogFile(self):
        log = Logging.Logger()
        self.assertTrue(log.get_logger(), Logger)
        del log

    def test_log_get_messages(self):
        log = Logging.Logger()
        log.debug('random message')
        self.assertEqual(1, len(log.get_messages()))
        del log

    def test_empty_logger(self):
        log = Logging.Logger()
        log.debug('random message')
        log.get_messages_dump()
        self.assertEqual(0, len(log.get_messages()))
        del log

    def test_log_file_write_line(self):
        log = Logging.Logger()
        # test for directory
        self.assertTrue(self._config.grease_dir + self._config.fs_Separator)
        # test log file exists
        self.assertTrue(self._config.grease_dir + self._config.fs_Separator + "grease.log")
        # test line count
        original = sum(1 for line in open(self._config.grease_dir + self._config.fs_Separator + "grease.log"))
        log.debug("I'm a test message")
        final = sum(1 for line in open(self._config.grease_dir + self._config.fs_Separator + "grease.log"))
        self.assertEqual(original + 1, final)



import sys
import time
from collections import deque
from tgt_grease_core.BaseCommand import GreaseCommand
from tgt_grease_core_util import Grease
from tgt_grease_core_util.ImportTools import Importer


class Router(object):
    """class Router handles routing requests to grease utilities"""
    def __init__(self):
        self.start_time = time.time()
        self.args = deque(sys.argv)
        # Pop the entrypoint off the stack
        self.args.popleft()
        self._grease = Grease()
        self._importer = Importer(self._grease.message())

    @staticmethod
    def entry_point():
        """entry_point is the 'main' method of the core system"""
        router = Router()
        router.gateway()

    def gateway(self):
        """gateway is the top level routing method for parsing commands"""
        if len(self.args) >= 2:
            instance = self._importer.load(self.args[0], self.args[1])
            if isinstance(instance, GreaseCommand):
                return_state = bool(instance.execute())
                Grease.run_telemetry(instance, return_state)
                if return_state:
                    self._grease.message().debug('Grease Request Successful::' + str(list(self.args)))
                    sys.exit(0)
                else:
                    self._grease.message().error('Grease Request Failed::' + str(list(self.args)))
                    self.bad_exit("Process Failed", 6)
            else:
                self.bad_exit("Loaded Command Does Not Implement GreaseCommand :: This is a development issue please "
                              "contact owner", 5)
        else:
            if len(self.args) == 1:
                self.bad_exit("No Subcommand Instruction Provided perhaps use help", 1)
            else:
                self.bad_exit("Subcommand Not Provided", 1)

    def bad_exit(self, message, status=0):
        # type: (str, int) -> None
        """bad_exit is for when the router needs to bail out"""
        self._grease.message().error(message + '::' + str(list(self.args)))
        print(str(message))
        sys.exit(int(status))



from unittest import TestCase
from tgt_grease_core_util import GreaseUtility
from tgt_grease_core_util import Logging

class test_greaseUtility(TestCase):
    def test_message(self):
        inst = GreaseUtility.Grease()
        self.assertEqual(type(inst.message()), type(Logging.Logger()))



from abc import ABCMeta, abstractmethod
from tgt_grease_core_util.GreaseUtility import Grease
import time


class GreaseCommand(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        self.purpose = "Abstract Base Class"
        self._machines_effected = 0
        self._start_time = time.time()
        self._ioc = Grease()

    def __del__(self):
        self._ioc.__del__()

    def set_effected(self, count):
        # type: (int) -> None
        self._machines_effected = int(count)

    def get_effected(self):
        # type: () -> int
        return int(self._machines_effected)

    def get_start_time(self):
        # type: () -> float
        return float(self._start_time)

    @abstractmethod
    def execute(self):
        """execute should be the primary logic for your execution"""
        pass



from . import Detectors
from .BaseSourceClass import BaseSource
from .Article14Section31 import Section31 as system_heartbeat
from .Detector import Detector as engage_lcars
from .Kirk import CaptainKirk as explore_galaxy
from .LaunchCtl import LaunchCtl as bridge
from .Scanner import ScanOnConfig as engage_sensor_array
from .Scheduler import Scheduler as engage_warp
from . import DeDuplification



from abc import ABCMeta, abstractmethod


class BaseDetector(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def param_compute(self, source, rules):
        # type: (dict, dict) -> None
        pass

    @abstractmethod
    def get_result(self):
        # type: () -> dict
        pass



from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand
from tgt_grease_enterprise.Detectors import DetectorConfiguration
from tgt_grease_core_util import SQLAlchemyConnection, Configuration
from tgt_grease_core_util.RDBMSTypes import JobServers, SourceData, JobConfig, JobQueue
from sqlalchemy import and_, update, or_
from datetime import datetime
import json
import hashlib


class Scheduler(GreaseDaemonCommand):
    def __init__(self):
        super(Scheduler, self).__init__()
        self._scanner_config = DetectorConfiguration.ConfigurationLoader()
        self._config = Configuration()
        self._sql = SQLAlchemyConnection(self._config)

    def execute(self, context='{}'):
        # Lets go get the jobs needing to be scheduled by this server
        result = self._sql.get_session().query(SourceData, JobServers)\
            .filter(SourceData.scheduling_server == JobServers.id)\
            .filter(JobServers.id == self._config.node_db_id())\
            .filter(SourceData.detection_start_time != None)\
            .filter(SourceData.detection_end_time != None)\
            .filter(SourceData.detection_complete == True)\
            .filter(SourceData.scheduling_start_time == None)\
            .filter(SourceData.scheduling_end_time == None)\
            .filter(SourceData.scheduling_complete == False)\
            .limit(15)\
            .all()
        if not result:
            self._ioc.message().debug("No Sources to schedule", True)
            return True
        else:
            for schedule in result:
                # lets own this
                self._take_ownership(schedule.SourceData.id)
                self._schedule_via_sources(schedule.SourceData.source_data)
                # lets finish out
                self._complete(schedule.SourceData.id)
        return True

    def _schedule_via_sources(self, sources):
        # type: (dict) -> None
        for index in sources:
            # check to see if we already assigned this source
            if 'grease_internal_assigned' in sources[index]:
                if sources[index]['grease_internal_assigned']:
                    # we already successfully did the thing
                    self._ioc.message().debug(
                        "SCHEDULING ALREADY PROCEEDED ON INDEX [{0}]".format(str(index)),
                        True
                    )
                    continue
                else:
                    # it failed previously
                    self._ioc.message().warning("Record Failed To Be Assigned::Loop Over")
                    continue
            else:
                # we have not attempted to schedule this record yet
                if len(sources[index]['rule_processing']) > 0:
                    # rules got processed on this source
                    for rule_name, rule_results in sources[index]['rule_processing'].iteritems():
                        self._ioc.message().debug(
                            "PROCESSING SOURCE [{0}] FOR RULE [{1}]".format(str(index), str(rule_name)),
                            True
                        )
                        if rule_results['status']:
                            self._ioc.message().debug(
                                "SOURCE [{0}] RULE [{1}] PASSED DETECTION".format(str(index), str(rule_name)),
                                True
                            )
                            # rule was successful time to schedule
                            # lets go load the rule config
                            rule_config = self._scanner_config.get_config(rule_name)
                            self._ioc.message().debug(
                                "READING CONFIG FOR RULE [{0}]".format(str(rule_name)),
                                True
                            )
                            # setup the execution environment variable
                            if 'exe_env' in rule_config:
                                if len(rule_config['exe_env']) > 0:
                                    # if we have a valid string then just set it
                                    exe_env = rule_config['exe_env']
                                else:
                                    # else they left it blank, default to general
                                    exe_env = 'general'
                            else:
                                # they didn't provide one so default to general
                                exe_env = 'general'
                            # next lets get the ID
                            if 'incident_number' in sources[index]:
                                # Set the incident Number
                                i_num = sources[index]['incident_number']
                            elif 'number' in sources[index]:
                                # Set the events number
                                i_num = sources[index]['number']
                            else:
                                # default to md5 hash of values list to ensure unique ID
                                i_num = hashlib.sha256(json.dumps(sources[index].values())).hexdigest()
                            # Now lets set the job additional parameters
                            additional = dict()
                            if 'parameters' in sources[index]['rule_processing'][rule_name]:
                                if isinstance(sources[index]['rule_processing'][rule_name]['parameters'], dict):
                                    additional = sources[index]['rule_processing'][rule_name]['parameters']
                                else:
                                    self._ioc.message().warning(
                                        "Parameters were not dictionary for rule: [" + str(rule_name) + "]"
                                    )
                            # Now lets setup the ticket info
                            additional['ticket'] = i_num
                            self._ioc.message().debug(
                                "PREPARING TO SCHEDULE JOB [{0}] FOR EXECUTION::"
                                "EXE_ENV: [{1}] ADDITIONAL: [{2}] ticket: [{3}]".format(
                                    str(rule_config['job']),
                                    str(exe_env),
                                    str(additional),
                                    str(i_num)
                                ),
                                True
                            )
                            if self._assign(
                                    rule_config['job'],
                                    exe_env,
                                    str(self._config.get('SCHEDULE_PKG', 'localhost_generic')),
                                    str(i_num),
                                    additional
                            ):
                                # we successfully assigned the ticket
                                self._ioc.message().debug(
                                    "JOB EXECUTION SCHEDULING SUCCESSFUL [{0}] FOR RULE [{1}]".format(
                                        str(index),
                                        str(rule_name)
                                    ),
                                    True
                                )
                                sources[index]['grease_internal_assigned'] = True
                                continue
                            else:
                                # we failed to assign the ticket
                                self._ioc.message().warning(
                                    "JOB EXECUTION SCHEDULING FAILED [{0}] FOR RULE [{1}]".format(
                                        str(index),
                                        str(rule_name)
                                    ),
                                    True
                                )
                                sources[index]['grease_internal_assigned'] = False
                                continue
                        else:
                            # rule failed fine then
                            self._ioc.message().debug(
                                "RULE FAILED FOR SOURCE [{0}] RULE [{1}]".format(
                                    str(index),
                                    str(rule_name)
                                ),
                                True
                            )
                            continue
                else:
                    # No rules were processed on this source
                    self._ioc.message().debug(
                        "RULE PROCESSING WAS EMPTY FOR SOURCE [{0}]".format(
                            str(index)
                        ),
                        True
                    )
                    continue

    def _assign(self, job, exec_env, package, ticket, additional=dict):
        # type: (str, str, str, str, dict) -> bool
        # check to ensure this ticket isn't already on the schedule
        self._ioc.message().debug(
            "Job Scheduling Starting::Job [{0}] Exec_Env [{1}] Package [{2}] Ticket [{3}] Additional [{4}]".format(
                str(job),
                str(exec_env),
                str(package),
                str(ticket),
                str(additional)
            ),
            True
        )
        if len(ticket) > 0:
            self._ioc.message().debug("Validating ticket ID not already in Job Queue", True)
            result = self._sql.get_session().query(JobQueue)\
                .filter(JobQueue.ticket == ticket)\
                .filter(or_(and_(JobQueue.in_progress == False, JobQueue.completed == False), JobQueue.in_progress == True))\
                .all()
            if result:
                self._ioc.message().warning("Ticket Already in Job Queue for Execution Ticket")
                return False
        # lets only get the least assigned server so we can round robin
        self._ioc.message().debug("Searching for Execution Server", True)
        result = self._sql.get_session()\
            .query(JobServers)\
            .filter(JobServers.execution_environment == exec_env)\
            .filter(JobServers.active == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if not result:
            self._ioc.message().error("No Execution Environments Found For Job: [" + job + "]", hipchat=True)
            return False
        server_info = result.id
        server_job_count = int(result.jobs_assigned)
        self._ioc.message().debug(
            "Job Server Selected [{0}] current assignment total [{1}]".format(server_info, server_job_count),
            True
        )
        self._ioc.message().debug("Searching for Job Configuration", True)
        result = self._sql.get_session().query(JobConfig)\
            .filter(JobConfig.command_module == package)\
            .filter(JobConfig.command_name == job)\
            .first()
        if not result:
            self._ioc.message().error(
                "No Jobs Configured For Requested Job: [" + job + "] for package: [" + package + "]", hipchat=True
            )
            return False
        job_id = result.id
        # Proceed to schedule
        self._ioc.message().debug("Creating new job in queue for job [{0}] on node [{1}]".format(job_id, server_info), True)
        JobToQueue = JobQueue(
            host_name=server_info,
            job_id=job_id,
            ticket=ticket,
            additional=additional
        )
        self._sql.get_session().add(JobToQueue)
        self._sql.get_session().commit()
        # that job and increment the assignment counter
        self._ioc.message().debug("incrementing jobs assigned on server [{0}]".format(server_info), True)
        self._sql.get_session().query(JobServers).filter_by(id=server_info).update({'jobs_assigned': server_job_count + 1})
        self._sql.get_session().commit()
        self._ioc.message().debug(
            "JOB [{0}] SCHEDULED FOR SERVER [{1}]".format(
                str(job_id),
                str(server_info)
            ),
            True
        )
        return True

    def _take_ownership(self, source_file_id):
        # type: (int) -> None
        stmt = update(SourceData)\
            .where(SourceData.id == source_file_id)\
            .values(scheduling_start_time=datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug(
            "TAKING OWNERSHIP OF SOURCE [{0}]".format(
                str(source_file_id),
            ),
            True
        )

    def _complete(self, source_file_id):
        # type: (int) -> None
        stmt = update(SourceData)\
            .where(SourceData.id == source_file_id)\
            .values(scheduling_complete=True, scheduling_end_time=datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        self._ioc.message().debug(
            "COMPLETING SOURCE [{0}]".format(
                str(source_file_id),
            ),
            True
        )



import os
import fnmatch
import json
import pkg_resources


class ConfigurationLoader(object):
    def __init__(self):
        self._raw_config = list()
        self._config_as_detectors = dict()
        self._scanners = list()
        self._load_config()

    def get_raw_config(self):
        # type: () -> list
        return self._raw_config

    def count_loaded_configurations(self):
        # type: () -> int
        return len(self.get_raw_config())

    def get_config_as_detectors(self):
        # type: () -> dict
        return self._config_as_detectors

    def get_scanners(self):
        # type: () -> list
        return self._scanners

    def get_scanner_config(self, detector_name):
        # type: (str) -> list
        if detector_name in self.get_config_as_detectors():
            return self.get_config_as_detectors()[detector_name]
        else:
            return list()

    def get_config(self, config_name):
        for conf in self.get_raw_config():
            if conf['name'] == config_name:
                return conf
        return {}

    def _load_config(self):
        # type: () -> None
        intermediate = list()
        dir_struct = pkg_resources.resource_filename(
            os.getenv('GREASE_CONF_PKG', ''),  os.getenv('GREASE_CONF_DIR', '') + '/'
        )
        matches = []
        for root, dirnames, filenames in os.walk(dir_struct):
            for filename in fnmatch.filter(filenames, '*.config.json'):
                matches.append(os.path.join(root, filename))
        result = matches
        for doc in result:
            with open(doc) as current_file:
                content = current_file.read().replace('\r\n', '')
            try:
                intermediate.append(json.loads(content))
            except ValueError:
                continue
        self._raw_config = intermediate
        result = {}
        for conf in self.get_raw_config():
            if conf['detector'] not in result:
                result[conf['detector']] = list()
            result[conf['detector']].append(conf)
        self._config_as_detectors = result
        result = list()
        for scanner, data in self.get_config_as_detectors().iteritems():
            result.append(scanner)
        self._scanners = result
        del result



from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand
from tgt_grease_core_util.Database import SQLAlchemyConnection
from tgt_grease_core_util import Configuration
from tgt_grease_core_util.RDBMSTypes import JobServers, JobQueue,SourceData, ServerHealth
from sqlalchemy import update, and_, or_
import datetime
import time


class Section31(GreaseDaemonCommand):
    def __init__(self):
        super(Section31, self).__init__()
        self._config = Configuration()
        self._sql = SQLAlchemyConnection(self._config)

    def execute(self, context='{}'):
        # first lets get the entire farm
        farm = self._get_farm_status()
        # now lets check each server from the last known state
        for server in farm:
            if self._server_alive(server):
                # server seems to be alive keep going
                continue
            else:
                # server is not alive, we need to migrate work from it
                # first lets declare ourselves as the doctor
                self._declare_doctor(server[0])
                # now lets sleep to ensure we don't stop on other doctors
                time.sleep(10)
                # okay lets ensure we are still the doctor
                if self._am_i_the_doctor(server[0]):
                    # time to reassign
                    self._cull_server(server[0])
                else:
                    # during our slumber someone else decided to work on this server let them have it
                    continue
        return True

    def _get_farm_status(self):
        # type: () -> list
        sql = """
            WITH on_demand AS (
                SELECT
                  jq.host_name,
                  count(jq.id) AS total
                FROM
                  job_queue jq
                WHERE
                  (
                    (
                      jq.in_progress = FALSE AND
                      jq.completed = FALSE
                    ) OR
                    jq.in_progress = TRUE
                  )
                GROUP BY jq.host_name
              ),
                completed_on_demand AS (
                  SELECT
                    jt.server_id as host_name,
                    count(jt.id) AS total
                  FROM
                    job_servers js
                    LEFT OUTER JOIN job_telemetry jt ON (jt.server_id = js.id)
                  GROUP BY jt.server_id
              ),
              completed_daemon AS (
                SELECT
                  js.id as host_name,
                  count(jtd.id) AS total
                FROM
                  job_servers js
                  LEFT OUTER JOIN job_telemetry_daemon jtd ON (jtd.server_id = js.id)
                GROUP BY js.id
              ),
                sources AS (
                  SELECT
                    js.id as host_name,
                    count(sf.id) AS total
                  FROM
                    source_data sf
                    INNER JOIN grease.job_servers js ON (js.id = sf.detection_server)
                  WHERE
                    (
                      (
                        sf.detection_start_time IS NOT NULL AND
                        sf.detection_end_time IS NOT NULL
                      ) OR
                      (
                        sf.detection_start_time IS NOT NULL AND
                        sf.detection_end_time IS NULL
                      )
                    )
                  GROUP BY js.id
              ),
                schedules AS (
                  SELECT
                    js.id as host_name,
                    count(sq.id) AS total
                  FROM
                    source_data sq
                    INNER JOIN grease.job_servers js ON (js.id = sq.scheduling_server)
                  WHERE
                    (
                      (
                        sq.scheduling_start_time IS NOT NULL AND
                        sq.scheduling_end_time IS NOT NULL
                      ) OR
                      (
                        sq.scheduling_start_time IS NOT NULL AND
                        sq.scheduling_end_time IS NULL
                      )
                    )
                  GROUP BY js.id
              )
            SELECT
              js.id                    AS job_server_id,
              js.host_name,
              coalesce(od.total, 0)    AS on_demand_total,
              coalesce((cod.total + coda.total), 0)   AS completed_total,
              coalesce(s.total, 0)     AS sources,
              coalesce(sch.total, 0)   AS schedules,
              (coalesce(od.total, 0) || '-' || coalesce((cod.total + coda.total), 0) || '-' || coalesce(s.total, 0) || '-' ||
               coalesce(sch.total, 0)) AS str_hash
            FROM
              job_servers js
              LEFT OUTER JOIN on_demand od ON (od.host_name = js.id)
              LEFT OUTER JOIN completed_on_demand cod ON (cod.host_name = js.id)
              LEFT OUTER JOIN completed_daemon coda ON (coda.host_name = js.id)
              LEFT OUTER JOIN sources s ON (s.host_name = js.id)
              LEFT OUTER JOIN schedules sch ON (sch.host_name = js.id)
            WHERE
              js.active IS TRUE
            ORDER BY js.id
        """
        return list(self._sql.get_engine().execute(sql).fetchall())

    def _declare_doctor(self, server_id):
        # type: (int) -> None
        # first lets check to ensure someone hasn't already claimed this beast
        result = self._sql.get_session().query(ServerHealth)\
            .filter(ServerHealth.server == server_id)\
            .first()
        if result and result.doctor != '':
            self._ioc.message().error("SERVER DOCTOR ALREADY DECLARED FOR [{0}]".format(server_id))
            return
        stmt = update(ServerHealth).where(ServerHealth.server == server_id).values(docter=self._config.node_db_id())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()

    def _am_i_the_doctor(self, server_id):
        # type: (int) -> bool
        result = self._sql.get_session().query(ServerHealth)\
            .filter(ServerHealth.doctor == self._config.node_db_id())\
            .filter(ServerHealth.server == server_id)\
            .first()
        if result:
            return True
        else:
            return False

    def _server_alive(self, server):
        # type: (list) -> bool
        result = self._sql.get_session().query(ServerHealth)\
            .filter(ServerHealth.server == server[0])\
            .first()
        if not result:
            # if this is a server not checked into health before
            self._register_in_health(server[0])
            result = self._sql.get_session().query(ServerHealth)\
                .filter(ServerHealth.server == server[0])\
                .first()
        # now return to regular logic
        if result.check_time < (datetime.datetime.now(result.check_time.tzinfo) - datetime.timedelta(hours=6)):
            # server status is old
            # lets check the hash though
            if result.job_hash == server[6]:
                # hash hasn't changed time for d-com
                return False
            else:
                # hash is old but changed, lets update and return True
                self._update_hash(server[0], server[6])
                return True
        else:
            # results aren't 12 hrs old yet
            if result.job_hash == server[6]:
                # log but do nothing so we know bad things will happen if we continue
                if result.check_time < (datetime.datetime.now(result.check_time.tzinfo) - datetime.timedelta(minutes=10)):
                    self._ioc.message().warning(
                        "Server hash has not changed since last poll ID:[{0}] hash:[{1}]".format(
                            server[0],
                            server[6]
                        )
                    )
            else:
                # if it changed lets update so we see the current hash
                self._update_hash(server[0], server[6])
            return True

    def _register_in_health(self, server):
        newServer = ServerHealth(
            server=server[0],
            job_hash=server[6],
            check_time=datetime.datetime.utcnow()
        )
        self._sql.get_session().add(newServer)
        self._sql.get_session().commit()

    def _update_hash(self, server_id, hash_str):
        # type: (str, str) -> None
        stmt = update(ServerHealth)\
            .where(ServerHealth.server == server_id)\
            .values(job_hash=hash_str, check_time=datetime.datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()

    def _deactivate_server(self, server_id):
        # type: (int) -> None
        stmt = update(JobServers)\
            .where(JobServers.id == server_id)\
            .values(active=False)
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()

    def _get_new_execution_server(self, server_id):
        # type: (int) -> JobServers
        # first get current env
        old_server = self._sql.get_session().query(JobServers)\
            .filter(JobServers.id == server_id)\
            .first()
        new_server = self._sql.get_session().query(JobServers)\
            .filter(JobServers.execution_environment == old_server.execution_environment)\
            .filter(JobServers.active == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if new_server:
            # we have a server to reassign to
            return new_server
        else:
            # there are no other execution servers for that environment
            self._ioc.message().error(
                "No other valid execution environment for jobs assigned to server [{0}]".format(server_id),
                hipchat=True
            )
            return False

    def _cull_server(self, server_id):
        # type: (int) -> None
        self._ioc.message().warning("DEACTIVATING SERVER [{0}]".format(server_id), hipchat=True)
        # first lets deactivate the job server
        self._deactivate_server(server_id)
        # next we need to check for any jobs scheduled to that instance and reassign them
        # first lets see if there is another available server
        server = self._get_new_execution_server(server_id)
        if server:
            # get the number of jobs being reassigned
            result = self._sql.get_session().query(JobQueue) \
                .filter(or_(and_(JobQueue.in_progress == False, JobQueue.completed == False), JobQueue.in_progress == True)) \
                .filter(JobQueue.host_name == server_id)\
                .all()
            if result:
                jobs = []
                for job in result:
                    jobs.append(job.id)
                # we have jobs to move
                # step one get the list of jobs to reassign
                job_id_group = tuple(jobs)
                self._reassign_on_demand_jobs(server_id, server.id, job_id_group)
            else:
                self._ioc.message().info("No jobs to reassign for server [{0}] being decommissioned".format(server_id))
        else:
            # no other valid server could be found for on-demand work
            self._ioc.message().error("Failed reassigning jobs for server [{0}]".format(server_id))
        # next lets move over any source detection work
        result = self._sql.get_session().query(SourceData)\
            .filter(SourceData.detection_server == server_id)\
            .filter(SourceData.detection_start_time == None)\
            .filter(SourceData.detection_end_time == None)\
            .filter(SourceData.detection_complete == False)\
            .all()
        jobs = []
        for job in result:
            jobs.append(job.id)
        source_file_jobs = tuple(jobs)
        if len(source_file_jobs):
            # we have sources to reassign parsing for
            self._reassign_sources(source_file_jobs, server_id)
        else:
            # no source documents to reassign
            self._ioc.message().info("No source parsing to reassign for server [{0}]".format(server_id))
        # finally lets worry about schedules to be reassigned
        result = self._sql.get_session().query(SourceData) \
            .filter(SourceData.scheduling_server == server_id) \
            .filter(SourceData.detection_start_time != None)\
            .filter(SourceData.detection_end_time != None)\
            .filter(SourceData.detection_complete == True)\
            .filter(SourceData.scheduling_start_time == None)\
            .filter(SourceData.scheduling_end_time == None)\
            .filter(SourceData.scheduling_complete == False)\
            .all()
        jobs = []
        for job in result:
            jobs.append(job.id)
        schedules = tuple(jobs)
        if len(schedules):
            # we have sources to reassign parsing for
            self._reassign_schedules(schedules, server_id)
        else:
            # no source documents to reassign
            self._ioc.message().info("No schedules to reassign for server [{0}]".format(server_id))

    def _reassign_on_demand_jobs(self, server_id, new_server, job_list):
        # type: (int, int, tuple) -> None
        # now lets do the update
        if not job_list:
            job_list = (-1,)
        stmt = update(JobQueue)\
            .where(JobQueue.id.in_(job_list))\
            .values(host_name=new_server)
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        # now lets update the totals on the job servers in question
        self._update_job_assignment_totals(server_id, new_server, len(job_list))
        self._ioc.message().info(
            "JOB ID SET [{0}] MOVED FROM SERVER [{1}] TO SERVER [{2}]".format(
                job_list,
                server_id,
                new_server
            )
        )

    def _reassign_sources(self, source_file_ids, origin_server):
        # type: (tuple, int) -> None
        # first lets find our applicable detector server
        result = self._sql.get_session().query(JobServers)\
            .filter(JobServers.active == True)\
            .filter(JobServers.detector == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if result:
            new_server = result.id
            reassign_total = len(source_file_ids)
            # Actual reassignment of sources
            stmt = update(SourceData)\
                .where(SourceData.id.in_(source_file_ids))\
                .values(detection_server=new_server)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            self._update_job_assignment_totals(origin_server, new_server, reassign_total)
        else:
            # oh crap we have no job detector servers left alive
            # Me IRL Right now:
            # https://68.media.tumblr.com/e59c51080e14cee56ce93416cf8055c8/tumblr_mzncp3nYXh1syplf0o1_250.gif
            self._ioc.message().error(
                "NO AVAILABLE DETECTOR SERVER CANNOT REASSIGN JOBS [{0}] FROM SERVER [{1}]".format(
                    source_file_ids, origin_server
                ),
                hipchat=True
            )
            # bail because we have no one to reassign to
            return

    def _reassign_schedules(self, schedule_ids, origin_server):
        # type: (tuple, int) -> None
        # first lets find our applicable scheduling server
        result = self._sql.get_session().query(JobServers)\
            .filter(JobServers.active == True)\
            .filter(JobServers.scheduler == True)\
            .order_by(JobServers.jobs_assigned)\
            .first()
        if result:
            new_server = result.id
            reassign_total = len(schedule_ids)
            # Actual reassignment of sources
            stmt = update(SourceData)\
                .where(SourceData.id.in_(schedule_ids))\
                .values(scheduling_server=new_server)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            self._update_job_assignment_totals(origin_server, new_server, reassign_total)
        else:
            # oh crap we have no job scheduler servers left alive
            # Me IRL Right now: https://media.giphy.com/media/HUkOv6BNWc1HO/giphy.gif
            self._ioc.message().error(
                "NO AVAILABLE SCHEDULER SERVER CANNOT REASSIGN JOBS [{0}] FROM SERVER [{1}]".format(
                    schedule_ids, origin_server
                ),
                hipchat=True
            )
            # bail because we have no one to reassign to
            return

    def _update_job_assignment_totals(self, old_server, new_server, job_count):
        # type: (int, int, int) -> None
        # first lets deduct from the bad server
        stmt1 = update(JobServers)\
            .where(id=old_server)\
            .values(jobs_assigned=JobServers.jobs_assigned - job_count)
        stmt2 = update(JobServers)\
            .where(id=new_server)\
            .values(jobs_assigned=JobServers.jobs_assigned + job_count)
        self._sql.get_session().execute(stmt1)
        self._sql.get_session().execute(stmt2)
        self._sql.get_session().commit()



from abc import ABCMeta, abstractmethod
import os
import sys
import atexit
import time
from signal import SIGTERM
from collections import deque
from tgt_grease_core_util import Configuration


class GreaseDaemonCommon(object):
    __metaclass__ = ABCMeta

    def __init__(self, router):
        # type: (DaemonRouter) -> None
        self._type = 'null'
        self._router = router
        self.args = deque(sys.argv)

    def get_type(self):
        # type: () -> str
        return str(self._type)

    def set_router(self, router):
        self._router = router

    def run(self):
        self._router.main()

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def restart(self):
        pass


class WindowsService(GreaseDaemonCommon):

    def __init__(self, args, router=None):
        # raise SystemError("Windows Daemons are not supported Yet")
        GreaseDaemonCommon.__init__(self, router)
        self.set_router(router)

    def restart(self):
        self.stop()
        self.start()

    def stop(self):
        sys.exit(0)

    def start(self):
        self._router.main()


class UnixDaemon(GreaseDaemonCommon):

    def __init__(self, router):
        super(UnixDaemon, self).__init__(router)
        self._config = Configuration()
        # self._pidfile = '/tmp/grease/grease.pid'
        self._pidfile = self._config.grease_dir + self._config.fs_Separator + 'grease.pid'
        self._pid = os.getpid()

    def daemonize(self):
        """Standard Double Fork"""
        # First Fork
        try:
            pid = os.fork()
            if pid > 0:
                # Exit Parent Now
                sys.exit(0)
        except OSError as e:
            self._router._log.critical("Fork Failure [1]: {0} ({1}) Host: [{2}]".format(
                e.errno,
                e.strerror,
                self._config.node_identity()
            ))
            self._router.bad_exit("Fork Failure: {0} ({1})".format(e.errno, e.strerror), 2)
        # Decouple from the parent process env data
        os.chdir("/")
        os.setsid()
        os.umask(0)
        # Second Fork
        try:
            pid = os.fork()
            if pid > 0:
                # Exit Parent Now
                sys.exit(0)
        except OSError as e:
            self._router._log.critical("Fork Failure [1]: {0} ({1}) Host: [{2}]".format(
                e.errno,
                e.strerror,
                self._config.node_identity()
            ))
            self._router.bad_exit("Fork Failure: {0} ({1})".format(e.errno, e.strerror), 2)
        # Finally lets write our pid file
        # Register Our Daemon
        atexit.register(self.del_pid)
        pid = str(os.getpid())
        file(self._pidfile, 'w+').write(str(pid))
        self._pid = pid

    def del_pid(self):
        os.remove(str(self._pidfile))

    def restart(self):
        self.stop()
        self.start()

    def stop(self):
        # Attempt to get lock on PID
        try:
            pidfile = file(self._pidfile, 'r')
            pid = int(pidfile.read().strip())
            pidfile.close()
        except IOError:
            pid = None
        # If there was no File / Pid Found
        if not pid:
            self._router._log.error("Daemon Not Running Cannot Stop")
            self._router.bad_exit("Daemon Is Not Running Currently, Failed To Stop", 3)
        # try to kill the running PID
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self._pidfile):
                    os.remove(self._pidfile)
            else:
                self._router._log.exception(err)
                self._router.bad_exit(err, 4)

    def start(self):
        # First check if daemon is already running
        try:
            pidfile = file(self._pidfile, 'r')
            pid = int(pidfile.read().strip())
            pidfile.close()
        except IOError:
            pid = None
        # if already in Daemon then ignore else start
        if pid:
            self._router._log.info("Daemon already running")
            print("Daemon Already In Service")
            sys.exit(0)
        else:
            self.daemonize()
            self.run()




import os
import pymongo
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


class SQLAlchemyConnection(object):
    """
        Provide access to PostgreSQL database created via SQLAlchemy
    """
    _config = None
    _engine = None
    _session = None

    def __init__(self, config):
        self._config = config
        if self._engine is None:
            self._engine = create_engine("{0}://{1}:{2}@{3}:{4}/{5}".format(
                self._config.get("GREASE_DB_ENGINE", "postgresql"),
                self._config.get("GREASE_DB_USER", "dev"),
                self._config.get("GREASE_DB_PASSWORD", "dev"),
                self._config.get("GREASE_DB_HOST", "localhost"),
                self._config.get("GREASE_DB_PORT", "5432"),
                self._config.get("GREASE_DB_DB", "")
            ))

    def get_engine(self):
        return self._engine

    def get_session(self):
        if not self._session:
            session = sessionmaker(bind=self.get_engine())
            self._session = session()
        return self._session


class MongoConnection(object):
    def __init__(self):
        if not os.getenv('GREASE_MONGO_USER') and not os.getenv('GREASE_MONGO_PASSWORD'):
            self._client = pymongo.MongoClient(
                host=os.getenv('GREASE_MONGO_HOST', 'localhost'),
                port=int(os.getenv('GREASE_MONGO_PORT', 27017))
            )
        else:
            self._client = pymongo.MongoClient(
                "mongodb://{0}:{1}@{2}:{3}/{4}".format(
                    os.getenv('GREASE_MONGO_USER', ''),
                    os.getenv('GREASE_MONGO_PASSWORD', ''),
                    os.getenv('GREASE_MONGO_HOST', 'localhost'),
                    os.getenv('GREASE_MONGO_PORT', 27017),
                    os.getenv('GREASE_MONGO_DB', 'grease')
                )
            )

    def client(self):
        # type: () -> pymongo.MongoClient
        return self._client



from abc import ABCMeta, abstractmethod
import time
import os
import threading
import sys
from tgt_grease_core.BaseCommand import GreaseCommand


class GreaseDaemonCommand(GreaseCommand):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(GreaseDaemonCommand, self).__init__()
        self.purpose = "Abstract Base Class"
        self._machines_effected = 0
        self._start_time = time.time()
        if os.name == 'nt':
            self._identity_file = "C:\\grease\\grease_identity.txt"
        else:
            self._identity_file = "/tmp/grease/grease_identity.txt"
        if os.path.isfile(self._identity_file):
            self._grease_identity = open(self._identity_file, 'r').read().rstrip()
        else:
            self._grease_identity = ''
        self.exe_state = {
            'execution': False,
            'result': False,
            'isDaemon': True
        }

    def set_effected(self, count):
        # type: (int) -> None
        self._machines_effected = int(count)

    def get_effected(self):
        # type: () -> int
        return int(self._machines_effected)

    def get_start_time(self):
        # type: () -> float
        return float(self._start_time)

    def attempt_execution(self, command_id, context='{}'):
        # type: (int, dict) -> None
        self.set_exe_state('command_id', int(command_id))
        try:
            self.set_exe_state('result', bool(self.execute(context)))
            self.set_exe_state('execution', True)
        except:
            self._ioc.message().error(
                "Command Execution Failed! Exception Raised: [{0}]".format(str(sys.exc_info())),
                hipchat=True
            )
            return
        return

    def get_exe_state(self):
        # type: () -> dict
        return self.exe_state

    def set_exe_state(self, key, value):
        # type: (str, object) -> None
        self.exe_state[key] = value

    @abstractmethod
    def execute(self, context='{}'):
        # type: (list) -> bool
        """execute should be the primary logic for your execution"""
        pass



from .BaseDetectorClass import BaseDetector


# Basic Range Detector Config
# ...
# "Range": [
#    {
#       "field": "my_field", <- Mandatory
#       "min": 10, <- Optional if max is used
#       "max": 100, <- Optional if min is used
#       "variable": True, <- Optional Assignment Parameter
#       "variable_name": "my_field" <- Optional Assignment Parameter
#    }
# ]
#
# ...


class Range(BaseDetector):
    def __init__(self):
        self._result = {'result': False}

    def param_compute(self, source, rules):
        rules_needed_to_pass = len(rules)
        rules_passed = 0
        for range_conf in rules:
            if range_conf['field'] in source:
                range_compare = False
                if 'max' in range_conf and 'min' in range_conf:
                    if int(range_conf['max']) < int(source[range_conf['field']]) \
                            or int(range_conf['min']) > int(source[range_conf['field']]):
                        range_compare = True
                elif 'min' in range_conf:
                    if int(range_conf['min']) > int(source[range_conf['field']]):
                        range_compare = True
                elif 'max' in range_conf:
                    if int(range_conf['max']) < int(source[range_conf['field']]):
                        range_compare = True
                else:
                    # invalid config no range parameters set
                    self._result = {'result': False}
                    return
                if range_compare:
                    rules_passed += 1
                    self._result['result'] = True
                    if 'variable' in range_conf:
                        if range_conf['variable'] and 'variable_name' in range_conf:
                            self._result[range_conf['variable_name']] = int(source[range_conf['field']])
            else:
                # field not found in source
                # time to fail out since not all range parameters can be met
                self._result = {'result': False}
                return
        if rules_needed_to_pass == rules_passed:
            return
        else:
            self._result = {'result': False}
            return

    def get_result(self):
        return self._result



from tgt_grease_daemon.BaseCommand import GreaseDaemonCommand
from .Article14Section31 import Section31
import os
import uuid
import sys
from tgt_grease_core_util import Configuration
from tgt_grease_core_util import SQLAlchemyConnection
from tgt_grease_core_util.RDBMSTypes import JobServers, ServerHealth, PersistentJobs, JobConfig, JobQueue
from datetime import datetime
from sqlalchemy import update, and_
from tgt_grease_core_util import RDBMSTypes


class LaunchCtl(GreaseDaemonCommand):

    _config = Configuration()
    _sql = SQLAlchemyConnection(_config)

    def __init__(self):
        super(LaunchCtl, self).__init__()
        self.purpose = "Register machine with Job Control Database"

    def execute(self, context='{}'):
        if len(sys.argv) >= 4:
            action = str(sys.argv[3])
        else:
            action = ''
        if action == 'register':
            return bool(self._action_register())
        elif action == 'kill-server':
            return bool(self._action_cull_server())
        elif action == 'revive-server':
            return bool(self._action_restore_server())
        elif action == 'list-pjobs':
            return bool(self._action_list_persistent_jobs())
        elif action == 'list-jobs':
            return bool(self._action_list_job_schedule())
        elif action == 'assign-task':
            return bool(self._action_assign_task())
        elif action == 'remove-task':
            return bool(self._action_remove_task())
        elif action == 'enable-detection':
            return bool(self._action_enable_detection())
        elif action == 'enable-scheduling':
            return bool(self._action_enable_scheduling())
        elif action == 'disable-detection':
            return bool(self._action_disable_detection())
        elif action == 'disable-scheduling':
            return bool(self._action_disable_scheduling())
        elif action == 'create-job':
            return bool(self._action_create_job())
        elif action == 'load-db':
            return bool(self._action_load_db())
        else:
            print("ERR: Invalid Command Expected: ")
            print("\tregister")
            print("\tkill-server")
            print("\trevive-server")
            print("\tlist-pjobs")
            print("\tlist-jobs")
            print("\tassign-task")
            print("\tremove-task")
            print("\tenable-detection")
            print("\tenable-scheduling")
            print("\tdisable-detection")
            print("\tdisable-scheduling")
            print("\tcreate-job")
            print("\tload-db")
            return True

    def _action_register(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            self._ioc.message().warning("Machine Already Registered With Grease Job Control")
            print("Machine Already Registered With Grease Job Control")
            return True
        else:
            # we need to register
            # first lets generate a new UUID
            uid = uuid.uuid4()
            # lets see if we have been provided an execution env
            if len(sys.argv) >= 5:
                exe_env = str(sys.argv[4])
            else:
                exe_env = 'general'
            # next lets register with the job control database
            server = JobServers(
                host_name=str(uid),
                execution_environment=exe_env,
                active=True,
                activation_time=datetime.utcnow()
            )
            self._sql.get_session().add(server)
            self._sql.get_session().commit()
            file(self._config.identity_file, 'w').write(str(uid))
            return True

    def _action_cull_server(self):
        # type: () -> bool
        if len(sys.argv) >= 5:
            server = str(sys.argv[4])
        else:
            if os.path.isfile(self._config.identity_file):
                server = self._config.identity
            else:
                print("Server has no registration record locally")
                return True
        # get the server ID
        result = self._sql.get_session().query(JobServers)\
            .filter(JobServers.host_name == server)\
            .first()
        if result:
            server_id = result.id
            instance = Section31()
            instance._declare_doctor(server_id)
            instance._cull_server(server_id)
            return True
        else:
            print("Job Server Not In Registry")
            return True

    def _action_restore_server(self):
        # type: () -> bool
        if len(sys.argv) >= 5:
            server = str(sys.argv[4])
        else:
            if os.path.isfile(self._config.identity_file):
                server = self._config.identity
            else:
                print("Server has no registration record locally")
                return True
        # get the server ID
        result = self._sql.get_session().query(JobServers)\
            .filter(JobServers.host_name == server)\
            .first()
        if result:
            server_id = result.id
        else:
            print("Job Server Not In Registry")
            return True
        # clear the doctor from the server health table
        stmt = update(ServerHealth)\
            .where(ServerHealth.server == server_id)\
            .values(doctor=None)
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        # next reactivate it
        stmt = update(JobServers)\
            .where(JobServers.id == server_id)\
            .values(active=True, activation_time=datetime.utcnow())
        self._sql.get_session().execute(stmt)
        self._sql.get_session().commit()
        return True

    def _action_list_persistent_jobs(self):
        # type: () -> bool
        result = self._sql.get_session().query(PersistentJobs, JobConfig)\
            .filter(PersistentJobs.command == JobConfig.id)\
            .filter(PersistentJobs.enabled == True)\
            .filter(PersistentJobs.server_id == self._config.node_db_id())\
            .all()
        if not result:
            print("No Scheduled Jobs on this node")
        else:
            for job in result:
                print(
                    "\tPackage: [{0}] Job: [{1}] Tick: [{2}] Additional: [{3}]".format(
                        job.JobConfig.command_module,
                        job.JobConfig.command_name,
                        job.JobConfig.tick,
                        job.PersistentJob.additional
                    )
                )
        return True

    def _action_list_job_schedule(self):
        # type: () -> bool
        result = self._sql.get_session().query(JobQueue)\
            .filter(JobQueue.completed == False)\
            .filter(JobQueue.in_progress == False)\
            .filter(JobQueue.host_name == self._config.node_db_id())\
            .all()
        if not result:
            print("No jobs scheduled on this node")
            return True
        for job in result:
            print("Jobs in Queue:")
            print("\t Module: [{0}] Command: [{1}] Additional: [{2}]".format(
                job.JobConfig.command_module,
                job.JobConfig.command_name,
                job.JobQueue.additional
            ))
        return True

    def _action_assign_task(self):
        # type: () -> bool
        if len(sys.argv) >= 5:
            new_task = str(sys.argv[4])
        else:
            print("Please provide a command to schedule to node")
            return True
        result = self._sql.get_session().query(JobConfig)\
            .filter(JobConfig.command_name == new_task)\
            .first()
        if not result:
            print("Command not found! Available Commands:")
            result = self._sql.get_session().query(JobConfig).all()
            if not result:
                print("NO JOBS CONFIGURED IN DB")
            else:
                for job in result:
                    print("{0}".format(job.command_name))
            return True
        else:
            pJob = PersistentJobs(
                server_id=self._config.node_db_id(),
                command=result.id,
                additional={},
                enabled=True
            )
            self._sql.get_session().add(pJob)
            self._sql.get_session().commit()
            print("TASK ASSIGNED")
            return True

    def _action_remove_task(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            server = self._config.node_db_id()
        else:
            print("Server has no registration record locally")
            return True
        if len(sys.argv) >= 5:
            new_task = str(sys.argv[4])
            result = self._sql.get_session().query(JobConfig).filter(JobConfig.command_name == new_task).first()
            if not result:
                print("Failed to find job in configuration tables")
                return True
            else:
                stmt = update(PersistentJobs)\
                    .where(and_(PersistentJobs.server_id == server, PersistentJobs.command == result.id))\
                    .values(enabled=False)
                self._sql.get_session().execute(stmt)
                self._sql.get_session().commit()
                print("TASK UNASSIGNED")
                return True

    def _action_enable_detection(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            server = self._config.identity
            stmt = update(JobServers).where(JobServers.host_name == server).values(detector=True)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            print("DETECTION ENABLED")
            return True
        else:
            print("ERR: SERVER UNREGISTERED")
            return False

    def _action_enable_scheduling(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            server = self._config.identity
            stmt = update(JobServers).where(JobServers.host_name == server).values(scheduler=True)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            print("SCHEDULING ENABLED")
            return True
        else:
            print("ERR: SERVER UNREGISTERED")
            return False

    def _action_disable_detection(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            server = self._config.identity
            stmt = update(JobServers).where(JobServers.host_name == server).values(detector=False)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            print("DETECTION DISABLED")
            return True
        else:
            print("ERR: SERVER UNREGISTERED")
            return False

    def _action_disable_scheduling(self):
        # type: () -> bool
        if os.path.isfile(self._config.identity_file):
            server = self._config.identity
            stmt = update(JobServers).where(JobServers.host_name == server).values(scheduler=False)
            self._sql.get_session().execute(stmt)
            self._sql.get_session().commit()
            print("SCHEDULING DISABLED")
            return True
        else:
            print("ERR: SERVER UNREGISTERED")
            return False

    def _action_create_job(self):
        # type: () -> bool
        if len(sys.argv) >= 6:
            new_task_module = str(sys.argv[4])
            new_task_class = str(sys.argv[5])
            # ensure a job doesn't exist
            result = self._sql.get_session().query(JobConfig)\
                .filter(JobConfig.command_module == new_task_module)\
                .filter(JobConfig.command_name == new_task_class)\
                .all()
            if result:
                print("Job already exists")
                return True
            NewJob = JobConfig(
                command_module=new_task_module,
                command_name=new_task_class
            )
            self._sql.get_session().add(NewJob)
            self._sql.get_session().commit()
            print("Job Created")
            return True
        else:
            print("ERR: INVALID SUBCOMMAND::MUST PASS MODULE AND CLASS NAME")
            return False

    def _action_load_db(self):
        print("LOADING DB")
        RDBMSTypes.__main__()
        if not os.path.isfile(self._config.identity_file):
            self._action_register()
        return True



import os
import time
import logging
from collections import deque
from logging.config import fileConfig
import random
from .Notifier import Notifier
from .Configuration import Configuration
# POSTGRES
from sqlalchemy.exc import OperationalError


class Logger:

    _config = Configuration()
    _unregisteredMode = False

    def __init__(self):
        self.start_time = time.time()
        self._messages = deque(())
        self._notifier = Notifier()
        if os.path.isfile(self._config.identity_file):
            try:
                self._node_id = self._config.node_db_id()
            except OperationalError:
                self._node_id = self._config.identity
        else:
            self._node_id = "UNREGISTERED"
        # Setup Log Configuration
        if type(self._config.get('GREASE_LOG_FILE')) == str and os.path.isfile(self._config.get('GREASE_LOG_FILE')):
            fileConfig(self._config.get('GREASE_LOG_FILE'))
            self._logger = logging.getLogger('GREASE-' + str(random.random()))
        else:
            logFilename = self._config.grease_dir + self._config.fs_Separator + "grease.log"
            self._logger = logging.getLogger('GREASE-' + str(random.random()))
            self._logger.setLevel(logging.DEBUG)
            self._handler = logging.FileHandler(logFilename)
            self._handler.setLevel(logging.DEBUG)
            self._formatter = logging.Formatter(
                "{\"timestamp\": \"%(asctime)s.%(msecs)03d\", \"node\": \""
                + str(self._node_id)
                + "\", \"thread\": \"%(threadName)s\", \"level\" : \"%(levelname)s\", \"message\" : \"%(message)s\"}",
                "%Y-%m-%d %H:%M:%S"
            )
            self._handler.setFormatter(self._formatter)
            self._logger.addHandler(self._handler)

    def __del__(self):
        self._logger.removeHandler(self._handler)

    def get_logger(self):
        # type: () -> logging
        return self._logger

    def get_messages(self):
        # type: () -> deque
        return self._messages

    def get_messages_dump(self):
        # type: () -> deque
        messages = self._messages
        self._messages = deque(())
        return messages

    def dress_message(self, message, level, hipchat, verbose, message_color='gray'):
        # type: (str, str, bool, bool, str) -> str
        message = "[{0}]::".format(str(self._node_id)) + message
        if verbose:
            message = "VERBOSE::" + message
        message = "{0}::".format(level) + message
        if hipchat:
            self._notifier.send_hipchat_message(message, message_color)
        return message

    def debug(self, message, verbose=False, hipchat=False):
        # type: (str, bool, bool) -> bool
        message = self.dress_message(message, "DEBUG", hipchat, verbose, 'gray')
        if verbose:
            if not self._config.get('GREASE_VERBOSE_LOGGING'):
                return True
        self._messages.append(('DEBUG', time.time(), message))
        return self._logger.debug(message)

    def info(self, message, verbose=False, hipchat=False):
        # type: (str, bool, bool) -> bool
        message = self.dress_message(message, "INFO", hipchat, verbose, 'purple')
        if verbose:
            if not self._config.get('GREASE_VERBOSE_LOGGING'):
                return True
        self._messages.append(('INFO', time.time(), message))
        return self._logger.info(message)

    def warning(self, message, verbose=False, hipchat=False):
        # type: (str, bool, bool) -> bool
        message = self.dress_message(message, "WARNING", hipchat, verbose, 'yellow')
        if verbose:
            if not self._config.get('GREASE_VERBOSE_LOGGING'):
                return True
        self._messages.append(('WARNING', time.time(), message))
        return self._logger.warning(message)

    def error(self, message, verbose=False, hipchat=False):
        # type: (str, bool, bool) -> bool
        message = self.dress_message(message, "ERROR", hipchat, verbose, 'red')
        if verbose:
            if not self._config.get('GREASE_VERBOSE_LOGGING'):
                return True
        self._messages.append(('ERROR', time.time(), message))
        return self._logger.error(message)

    def critical(self, message):
        # type: (str) -> bool
        message = self.dress_message(message, "CRITICAL", True, False, 'red')
        self._messages.append(('CRITICAL', time.time(), message))
        return self._logger.critical(message)

    def exception(self, message):
        # type: (str) -> bool
        message = self.dress_message(message, "EXCEPTION", True, False, 'red')
        self._messages.append(('EXCEPTION', time.time(), message))
        return self._logger.exception(message)



from tgt_grease_core import GreaseRouter
from tgt_grease_daemon import GreaseDaemonCommand
from tgt_grease_core_util import Configuration
from tgt_grease_core_util import Importer
from tgt_grease_core_util import Logger
from tgt_grease_core_util.RDBMSTypes import JobQueue, PersistentJobs, JobConfig
from tgt_grease_core_util import SQLAlchemyConnection
from sqlalchemy import update, and_, or_
from datetime import datetime
from tgt_grease_core_util import Grease
from . import Daemon
import threading
import os
import sys

# PyWin32
if os.name == 'nt':
    import win32event


class DaemonRouter(GreaseRouter.Router):
    """
    Daemon process routing for GREASE
    """
    __author__ = "James E. Bell Jr"
    __version__ = "1.0"

    _config = Configuration()
    _runs = 0
    _throttle_tick = 0
    _job_completed_queue = []
    _current_real_second = datetime.now().second
    _current_run_second = datetime.now().second
    _log = Logger()
    _process = None
    _importer = Importer(_log)
    _job_metadata = {'normal': 0, 'persistent': 0}
    _alchemyConnection = SQLAlchemyConnection(_config)
    _ioc = Grease()
    _ContextMgr = []

    def __init__(self):
        GreaseRouter.Router.__init__(self)
        if len(self._config.identity) <= 0:
            # ensure we won't run without proper registration
            print("ERR::Unregistered to Database!")
            self._log.error("Registration not found")
            sys.exit(1)

    @staticmethod
    def entry_point():
        """
        Application Entry point
        :return: void
        """
        router = DaemonRouter()
        router.gateway()

    def gateway(self):
        if len(self.args) >= 1:
            if self._config.op_name == 'nt':
                self.set_process(Daemon.WindowsService(sys.argv, self))
            else:
                self.set_process(Daemon.UnixDaemon(self))
            if self.args[0] == 'start':
                self._log.debug("Starting Daemon")
                self.get_process().start()
            elif self.args[0] == 'restart':
                self._log.debug("Restarting Daemon")
                self.get_process().restart()
            elif self.args[0] == 'stop':
                self._log.debug("Stopping Daemon")
                self.get_process().stop()
            else:
                self.bad_exit("Invalid Command To Daemon expected [start,stop,restart]", 2)
        else:
            self.bad_exit("Command not given to daemon! Expected: [start,stop,restart]", 1)

    def main(self):
        """
        Main Daemon Method
        :return: void
        """
        # Job Execution
        self._log.debug("PROCESS STARTUP", True)
        # initial rc value
        rc = "Garbage"
        # All programs are just loops
        if self._config.get('GREASE_EXECUTE_LINEAR'):
            self._log.debug("LINEAR EXECUTION MODE DETECTED")
        while True:
            # Windows Signal Catching
            if self._config.op_name == 'nt':
                if not rc != win32event.WAIT_OBJECT_0:
                    self._log.debug("Windows Kill Signal Detected! Closing GREASE")
                    break
            # Continue Processing
            # Process Throttling
            if self._config.get('GREASE_THROTTLE'):
                if int(self.get_throttle_tick()) > int(str(self._config.get('GREASE_THROTTLE'))):
                    # prevent more than 1000 loops per second by default
                    # check time
                    self.log_message_once_a_second("Throttle reached", -11)
                    self.have_we_moved_forward_in_time()
                    continue
            # Job Processing
            if self._config.get('GREASE_EXECUTE_LINEAR'):
                self.process_queue_standard()
            else:
                self.process_queue_threaded()
            # Final Clean Up
            self.inc_runs()
            self.inc_throttle_tick()
            self.have_we_moved_forward_in_time()
            # After all this check for new windows services
            if os.name == 'nt':
                # Block .5ms to listen for exit sig
                rc = win32event.WaitForSingleObject(self.get_process().hWaitStop, 50)

    def log_message_once_a_second(self, message, queue_id):
        # type: (str, int) -> bool
        # have we moved forward since the last second
        if self.have_we_moved_forward_in_time():
            # if we have we can log since the log does not have a record for this second
            self._log.debug(message)
            # We also ensure we record we already logged for zero jobs to process this second
            self.add_job_to_completed_queue(queue_id)
            return True
        else:
            # we have not moved forward in time
            # have we logged for this second
            if not self.has_job_run(queue_id):
                # We have not logged for this second so lets do that now
                self._log.debug(message)
                # record that we logged for this second
                self.add_job_to_completed_queue(queue_id)
                return True
            else:
                return False

    def process_queue_standard(self):
        # type: () -> bool
        job_queue = self.get_assigned_jobs()
        if len(job_queue) is 0:
            self.log_message_once_a_second("Total Jobs To Process: [0] Current Runs: [{0}]".format(self.get_runs()), -1)
        else:
            # We have some jobs to process
            if self._job_metadata['normal'] > 0:
                # if we have any normal jobs lets log
                self._log.debug("Total Jobs To Process: [{0}] Current Runs: [{1}]".format(
                    self._job_metadata['normal'],
                    self.get_runs()
                )
                )
            else:
                # we only have persistent jobs to process
                self.log_message_once_a_second("Total Jobs To Process: [{0}] Current Runs: [{1}]".format(
                        len(job_queue),
                        self.get_runs()
                    ),
                    0
                )
            # now lets loop through the job schedule
            for job in job_queue:
                # start class up
                command = self._importer.load(job['module'], job['command'])
                # ensure we got back the correct type
                if not command:
                    self._log.error(
                        "Failed To Load Command [{0}] of [{1}]"
                        .format(
                            job['command'],
                            job['module']
                        ),
                        hipchat=True
                    )
                    del command
                    continue
                if not isinstance(command, GreaseDaemonCommand):
                    self._log.error("Instance created was not of type GreaseDaemonCommand", hipchat=True)
                    del command
                    continue
                if not job['persistent']:
                    # This is an on-demand job
                    # we just need to execute it
                    self.mark_job_in_progress(job['id'])
                    self._log.debug("Preparing to execute on-demand job [{0}]".format(job['id']), True)
                    command.attempt_execution(job['id'], job['additional'])
                else:
                    # This is a persistent job
                    if self.has_job_run(job['id']):
                        # Job Already Executed
                        continue
                    else:
                        if job['tick'] is self.get_current_run_second():
                            self._log.debug("Preparing to execute persistent job [{0}]".format(job['id']), True)
                            command.attempt_execution(job['id'], job['additional'])
                            self.add_job_to_completed_queue(job['id'])
                        else:
                            # continue because we are not in the tick required
                            continue
                # Report Telemetry
                self._ioc.run_daemon_telemetry(command)
                if command.get_exe_state()['result']:
                    # job success
                    if job['persistent']:
                        self._log.debug("Persistent Job Successful [{0}]".format(job['id']))
                    else:
                        self._log.debug("On-Demand Job Successful [{0}]".format(job['id']))
                        self.mark_job_complete(job['id'])
                else:
                    # job failed
                    if job['persistent']:
                        self._log.debug("Persistent Job Failed [{0}]".format(job['id']))
                    else:
                        self._log.debug("On-Demand Job Failed [{0}]".format(job['id']))
                        self.mark_job_failure(job['id'], job['failures'])
                command.__del__()
                del command
        return True

    def process_queue_threaded(self):
        # type: () -> bool
        self.thread_check()
        job_queue = self.get_assigned_jobs()
        if len(job_queue) is 0:
            # have we moved forward since the last second
            self.log_message_once_a_second("Total Jobs To Process: [0] Current Runs: [{0}]".format(self.get_runs()), -1)
        else:
            if len(self._ContextMgr) >= int(self._config.get('GREASE_THREAD_MAX', '15')):
                self.log_message_once_a_second("Thread Maximum Reached::Current Runs: [{0}]".format(
                            self.get_runs()
                        ),
                    -10
                )
                return True
            # We have some jobs to process
            if self._job_metadata['normal'] is 0:
                # we only have persistent jobs to process
                self.log_message_once_a_second("Total Jobs To Process: [{0}] Current Runs: [{1}]".format(
                            len(job_queue),
                            self.get_runs()
                    ),
                    0
                )
            else:
                self._log.debug("Total Jobs To Process: [0] Current Runs: [{0}]".format(self.get_runs()), -1)
            # now lets loop through the job schedule
            for job in job_queue:
                # start class up
                command = self._importer.load(job['module'], job['command'])
                # ensure we got back the correct type
                if not command:
                    self._log.error(
                        "Failed To Load Command [{0}] of [{1}]"
                        .format(
                            job['command'],
                            job['module']
                        ),
                        hipchat=True
                    )
                    del command
                    continue
                if not isinstance(command, GreaseDaemonCommand):
                    self._log.error("Instance created was not of type GreaseDaemonCommand", hipchat=True)
                    del command
                    continue
                if not job['persistent']:
                    # This is an on-demand job
                    # we just need to execute it
                    self.mark_job_in_progress(job['id'])
                    self._log.debug("Passing on-demand job [{0}] to thread manager".format(job['id']), True)
                    self.thread_execute(command, job['id'], job['additional'], False, job['failures'])
                else:
                    # This is a persistent job
                    if self.has_job_run(job['id']):
                        # Job Already Executed
                        command.__del__()
                        del command
                        continue
                    else:
                        if job['tick'] is self.get_current_run_second():
                            self._log.debug("Passing persistent job [{0}] to thread manager".format(job['id']), True)
                            self.thread_execute(command, job['id'], job['additional'], True)
                            self.add_job_to_completed_queue(job['id'])
                        else:
                            # continue because we are not in the tick required
                            command.__del__()
                            del command
                            continue
        self.thread_check()
        return True

    def thread_execute(self, command, cid, additional, persistent, failures=0):
        # type: (GreaseDaemonCommand, int, dict, bool, int) -> None
        # first ensure the command ID isn't already running
        process_running = False
        for item in self._ContextMgr:
            if item[2] == cid:
                process_running = True
                break
        if process_running:
            # if it is return out
            self._log.debug("Bailing on job [{0}], already executing".format(cid), True)
            return None
        # start process
        proc = threading.Thread(
                target=command.attempt_execution,
                args=(cid, additional),
                name="GREASE EXECUTION THREAD::CID [{0}]".format(cid)
            )
        # set for background
        proc.daemon = True
        if persistent:
            self._log.debug("Beginning persistent execution of job [{0}] on thread".format(cid), True)
        else:
            self._log.debug("Beginning on-demand execution of job [{0}] on thread".format(cid), True)
        # start
        proc.start()
        # add command to pool
        self._ContextMgr.append([
            command,
            proc,
            cid,
            persistent,
            failures
        ])
        return None

    def thread_check(self):
        final = []
        # Check for tread completion else add back to list
        for command in self._ContextMgr:
            if command[1].isAlive():
                final.append(command)
            else:
                self._log.info("Job completed [{0}]".format(command[2]), True)
                self.record_telemetry(command[0], command[2], command[4], command[3])
        self._ContextMgr = final
        return

    def record_telemetry(self, command, cid, failures, persistent):
        # type: (GreaseDaemonCommand, int, int, bool) -> None
        # Report Telemetry
        command.set_exe_state('command_id', cid)
        self._ioc.run_daemon_telemetry(command)
        if command.get_exe_state()['result']:
            # job success
            if persistent:
                self._log.debug("Persistent Job Successful [{0}]".format(cid))
            else:
                self._log.debug("On-Demand Job Successful [{0}]".format(cid))
                self.mark_job_complete(cid)
        else:
            # job failed
            if persistent:
                self._log.debug("Persistent Job Failed [{0}]".format(cid))
            else:
                self._log.debug("On-Demand Job Failed [{0}]".format(cid))
                self.mark_job_failure(cid, failures)
        command.__del__()
        del command
        pass

    def mark_job_in_progress(self, job_id):
        """
        sets job as in progress
        :param job_id: int
        :return: bool
        """
        stmt = update(JobQueue).where(JobQueue.id == job_id).values(in_progress=True, completed=False)
        self._alchemyConnection.get_session().execute(stmt)
        self._alchemyConnection.get_session().commit()
        return True

    def mark_job_complete(self, job_id):
        """
        Complete a successful job
        :param job_id: int
        :return: bool
        """
        stmt = update(JobQueue).where(JobQueue.id == job_id).values(in_progress=False, completed=True, complete_time=datetime.now())
        self._alchemyConnection.get_session().execute(stmt)
        self._alchemyConnection.get_session().commit()
        return True

    def mark_job_failure(self, job_id, current_failures):
        """
        Fail a job
        :param job_id: int
        :param current_failures: int
        :return: bool
        """
        stmt = update(JobQueue)\
            .where(JobQueue.id == job_id)\
            .values(in_progress=False, completed=False, complete_time=None, failures=current_failures + 1)
        self._alchemyConnection.get_session().execute(stmt)
        self._alchemyConnection.get_session().commit()
        return True

    def get_assigned_jobs(self):
        """
        gets current job assignment
        :return: list
        """
        # type: () -> list
        # reset job queue metadata
        self._job_metadata['normal'] = 0
        self._job_metadata['persistent'] = 0
        # create final result
        final = []
        # first find normal jobs
        result = self._alchemyConnection\
            .get_session()\
            .query(JobQueue, JobConfig)\
            .filter(JobQueue.host_name == self._config.node_db_id())\
            .filter(JobQueue.job_id == JobConfig.id) \
            .filter(or_(and_(JobQueue.in_progress == False, JobQueue.completed == False), JobQueue.in_progress == True)) \
            .filter(JobQueue.failures < 6)\
            .all()
        if not result:
            # No Jobs Found
            self._job_metadata['normal'] = 0
        else:
            # Walk the job list
            for job in result:
                self._job_metadata['normal'] += 1
                final.append({
                    'id': job.JobQueue.id,
                    'module': job.JobConfig.command_module,
                    'command': job.JobConfig.command_name,
                    'request_time': datetime.utcnow(),
                    'additional': job.JobQueue.additional,
                    'tick': job.JobConfig.tick,
                    'persistent': False,
                    'failures': job.JobQueue.failures
                })
        # Now search for persistent jobs
        result = self._alchemyConnection\
            .get_session()\
            .query(PersistentJobs, JobConfig)\
            .filter(PersistentJobs.server_id == self._config.node_db_id())\
            .filter(PersistentJobs.command == JobConfig.id)\
            .filter(PersistentJobs.enabled == True)\
            .all()
        if not result:
            # No Jobs Found
            self._job_metadata['persistent'] = 0
        else:
            # Walk the job list
            for job in result:
                self._job_metadata['persistent'] += 1
                final.append({
                    'id': job.PersistentJobs.id,
                    'module': job.JobConfig.command_module,
                    'command': job.JobConfig.command_name,
                    'request_time': datetime.utcnow(),
                    'additional': job.PersistentJobs.additional,
                    'tick': job.JobConfig.tick,
                    'persistent': True,
                    'failures': 0
                })
        return final

    # Class Property getter/setters/methods

    # run counter
    def get_runs(self):
        """
        returns int of amount of loops
        :return: int
        """
        # type: () -> int
        return int(self._runs)

    def inc_runs(self):
        """
        increments run count
        :return: None
        """
        # type: () -> bool
        self._runs += 1
        return True

    def reset_runs(self):
        """
        resets the run counter to 0
        :return: bool
        """
        # type: () -> bool
        self._runs = 0
        return True

    # Job Completed Queue
    def add_job_to_completed_queue(self, job_ib):
        """
        Adds Job to queue so we don't run the job again
        :param job_ib: int
        :return: bool
        """
        # type: int -> bool
        if int(job_ib) not in self._job_completed_queue:
            self._log.debug("Job Executed This Second [{0}]".format(job_ib), True)
            self._job_completed_queue.append(int(job_ib))
            return True
        else:
            return False

    def has_job_run(self, job_id):
        """
        Determines if the job ID has run during the current cycle
        :param job_id: int
        :return: bool
        """
        # type: int -> bool
        if int(job_id) in self._job_completed_queue:
            return True
        else:
            return False

    def reset_completed_job_queue(self):
        """
        clears job run queue
        :return: bool
        """
        # type: () -> bool
        self._log.debug("Completed Per-Second Queue Cleared", True)
        self._job_completed_queue = []

    # throttle tick
    def get_throttle_tick(self):
        """
        returns how many runs in this second
        :return: int
        """
        # type: () -> int
        return int(self._throttle_tick)

    def inc_throttle_tick(self):
        """
        increases throttle tick by 1
        :return: bool
        """
        # type: () -> bool
        self._throttle_tick += 1
        return True

    def reset_throttle_tick(self):
        """
        resets throttle tick to 0
        :return: bool
        """
        # type: () -> bool
        self._throttle_tick = 0
        return True

    # Process Controller
    def set_process(self, process):
        """
        sets the process handler
        :param process: Daemon
        :return: None
        """
        self._process = process
        return None

    def get_process(self):
        """
        Returns _process property
        :return: Daemon/None
        """
        return self._process

    # time operators
    @staticmethod
    def get_current_real_second():
        """
        Gets current second
        :return: int
        """
        # type: () -> int
        return datetime.now().second

    def get_current_run_second(self):
        """
        Gets the current observed second
        :return: int
        """
        # type: () -> int
        return int(self._current_run_second)

    def set_current_run_second(self, sec):
        """
        Sets current observed second
        :param sec: int
        :return: None
        """
        # type: int -> None
        self._current_run_second = int(sec)

    def have_we_moved_forward_in_time(self):
        """
        Answers the question "have we moved forward in time?"
        so we reset our counters and return true else false
        :return: bool
        """
        # type: () -> bool
        if self.get_current_run_second() == self.get_current_real_second():
            return False
        else:
            self._log.debug("Time has moved forward! Restoring Context", True)
            self.set_current_run_second(self.get_current_real_second())
            self.reset_completed_job_queue()
            self.reset_throttle_tick()
            return True



from tgt_grease_core_util import Logging
from tgt_grease_core_util.Database import MongoConnection
import datetime
from pymongo.errors import ServerSelectionTimeoutError
import json
import hashlib
import os
import pymongo
from difflib import SequenceMatcher


class SourceDeDuplify(object):
    def __init__(self, logger):
        # type: (Logging.Logger) -> None
        self._logger = logger
        try:
            self._mongo_connection = MongoConnection()
            self._client = self._mongo_connection.client()
            self._db = self._client.get_database(
                name=os.getenv('GREASE_MONGO_DB', 'grease'),
                write_concern=pymongo.WriteConcern(w=0)
            )
            self._collection = self._db.get_collection(name='source_dedup')
            self._dedup = True
        except ServerSelectionTimeoutError:
            self._mongo_connection = None
            self._client = None
            self._db = None
            self._collection = None
            self._dedup = False

    def create_unique_source(self, source_name, field_set, source):
        # type: (str, list, list) -> list
        if not self._dedup:
            self._logger.error("Failed To connect to MongoDB de-duplication is off", hipchat=True)
            return source
        # we could connect lets process
        if not isinstance(source, list):
            self._logger.error('DeDuplification did not receive list returning empty', hipchat=True)
            return []
        self._logger.debug("STARTING DEDUPLICATION::TOTAL OBJECTS TO PROCESS [" + str(len(source)) + "]")
        final = []
        # now comes James' version of machine learning. I call it "Blue Collar Machine Learning"
        source_pointer = 0
        source_max = len(source)
        for source_obj in source:
            if not isinstance(source_obj, dict):
                self._logger.warning('DeDuplification Received NON-DICT Type: [' + str(type(source_obj)) + ']')
                continue
            source_pointer += 1
            # first thing try to find the object level hash
            hash_obj = self._collection.find_one({'hash': self.generate_hash(source_obj)})
            if not hash_obj:
                self._logger.debug("Failed To Locate Type1 Match, Performing Type2 Search Match", True)
                # Globally unique hash for request
                # create a completely new document hash and all the field set hashes
                self._collection.insert_one({
                    'expiry': self.generate_expiry_time(),
                    'max_expiry': self.generate_max_expiry_time(),
                    'score': 1,
                    'hash': self.generate_hash(source_obj)
                })
                # Next start field level processing
                # first check if our fields are limited
                if len(field_set) < 1:
                    # All fields need to be considered for de-dup
                    fields = source_obj.keys()
                else:
                    # only registered fields
                    fields = field_set
                # now lets get the composite score
                composite_score = self.get_field_score(source_name, source_obj, fields)
                self._logger.debug(
                    "DEDUPLICATION COMPOSITE SCORE ["
                    + str(source_pointer)
                    + "/"
                    + str(source_max)
                    + "]: "
                    + str(composite_score)
                )
                # now lets observe to see if we have a 'unique' source
                if composite_score < float(os.getenv('GREASE_DEDUP_SCORE', 85)):
                    # look at that its time to add it to the final list
                    self._logger.debug("Type2 ruled Unique adding to final result", True)
                    final.append(source_obj)
            else:
                # we have a duplicate source document
                # increase the counter and expiry and move on (DROP)
                self._logger.debug("Type1 match found, dropping", True)
                if 'max_expiry' in hash_obj:
                    update_statement = {
                        "$set": {
                            'score': int(hash_obj['score']) + 1,
                            'expiry': self.generate_expiry_time()
                        }
                    }
                else:
                    update_statement = {
                        "$set": {
                            'score': int(hash_obj['score']) + 1,
                            'expiry': self.generate_expiry_time(),
                            'max_expiry': self.generate_max_expiry_time()
                        }
                    }
                self._collection.update_one(
                    {'_id': hash_obj['_id']},
                    update_statement
                )
        # create the auto del index if it doesnt already exist
        self._collection.create_index([('expiry', 1), ('expireAfterSeconds', 1)])
        self._collection.create_index([('max_expiry', 1), ('expireAfterSeconds', 1)])
        remaining = len(final) - 1
        if remaining < 0:
            remaining = 0
        self._logger.debug("DEDUPLICATION COMPLETE::REMAINING OBJECTS [" + str(remaining) + "]")
        return final

    def get_field_score(self, source_name, document, field_set):
        # type: (str, dict, list) -> float
        # This function does a field by field assessment of a source
        # Woo is what I said too
        # first lets create our array of field scores
        score_list = []
        for field in field_set:
            if field in document:
                # unicode check because idk even man + Sha256 hash if too long
                if not document[field]:
                    document[field] = ''
                if isinstance(document[field], unicode):
                    check_document = {
                        'source': str(source_name),
                        'field': str(field),
                        'value': document[field].encode('utf-8')
                    }
                else:
                    check_document = {
                        'source': str(source_name),
                        'field': str(field),
                        'value': str(document[field]).encode('utf-8')
                    }
                # lets only observe fields in our list
                field_obj = self._collection.find_one({'hash': self.generate_hash(check_document)})
                if not field_obj:
                    self._logger.debug("Failed To Locate Type2 Match, Performing Type2 Search", True)
                    # globally unique field->value pair
                    # alright lets start the search for all fields of its name
                    field_probability_list = []
                    for record in self._collection.find({'source': str(source_name), 'field': field})\
                            .sort('score', pymongo.ASCENDING) \
                            .limit(int(os.getenv('GREASE_DEDUP_INSPECTION_STRENGTH', 100))):
                        # Now we are looping through our fields
                        if self.compare_strings(record['value'], check_document['value']) > .95:
                            # we found a VERY strong match
                            score_list.append(1 - self.compare_strings(record['value'], check_document['value']))
                            # leave the for loop since we found a highly probable match
                            break
                        else:
                            field_probability_list.append(self.compare_strings(record['value'], check_document['value']))
                    self._logger.debug("Failed To Location Type1 Match, Performing Type2 Search", True)
                    # lets record it in the 'brain' to remember it
                    check_document['hash'] = self.generate_hash(check_document)
                    check_document['score'] = 1
                    check_document['expiry'] = self.generate_expiry_time()
                    check_document['max_expiry'] = self.generate_max_expiry_time()
                    self._collection.insert_one(check_document)
                    # now lets either choose the highest probable match we found or 0 being a completely globally
                    # unique value (No matches found in the above loop
                    if len(field_probability_list) > 0:
                        probability_average = float(sum(field_probability_list) / len(field_probability_list))
                        self._logger.debug("Field Probability Average: source [{0}] field [{1}] is [{2}]"
                                           .format(source_name, field, probability_average), True)
                        score_list.append(probability_average)
                    else:
                        score_list.append(100)
                else:
                    self._logger.debug("Found Type2 Match! Dropping", True)
                    # exact match we can bug out
                    if 'max_expiry' in field_obj:
                        update_statement = {
                            "$set": {
                                'score': int(field_obj['score']) + 1,
                                'expiry': self.generate_expiry_time()
                            }
                        }
                    else:
                        update_statement = {
                            "$set": {
                                'score': int(field_obj['score']) + 1,
                                'expiry': self.generate_expiry_time(),
                                'max_expiry': self.generate_max_expiry_time()
                            }
                        }
                    self._collection.update_one(
                        {'_id': field_obj['_id']},
                        update_statement
                    )
                    score_list.append(0)

        # finally return our aggregate field score
        if len(score_list) is 0:
            return 0
        return sum(score_list) / float(len(score_list))

    @staticmethod
    def generate_hash(document):
        # type: (dict) -> str
        return hashlib.sha256(json.dumps(document)).hexdigest()

    @staticmethod
    def generate_expiry_time():
        return datetime.datetime.utcnow() + datetime.timedelta(hours=int(os.getenv('GREASE_DEDUP_TIMEOUT', 12)))

    @staticmethod
    def generate_max_expiry_time():
        return datetime.datetime.utcnow() + datetime.timedelta(weeks=int(os.getenv('GREASE_DEDUP_MAX_TIMEOUT', 1)))

    @staticmethod
    def compare_strings(constant, variable):
        # type: (str, str) -> float
        return SequenceMatcher(constant, variable).ratio()



from .Logging import Logger
from .Telemetry import Database, DatabaseDaemon


class Grease:
    def __init__(self):
        self._logger = Logger()

    def __del__(self):
        self._logger.__del__()

    def message(self):
        # type: () -> Logging.Logger
        return self._logger

    @staticmethod
    def run_telemetry(command_obj, run_success, is_daemon=False):
        # type: (tgt_grease_core.GreaseCommand, bool, bool) -> None
        telemetry = Database(command_obj, is_daemon)
        return telemetry.store_result(run_success)

    @staticmethod
    def run_daemon_telemetry(command_obj):
        # type: (tgt_grease_daemon.GreaseDaemonCommand) -> None
        telemetry = DatabaseDaemon(command_obj)
        return telemetry.store_result()



from .GreaseUtility import Grease
from .Database import MongoConnection, SQLAlchemyConnection
from .ImportTools import Importer
from .Notifier import Notifier
from .Logging import Logger
from . import RDBMSTypes
from .Configuration import Configuration



from setuptools import setup, find_packages
import os


setup(
    name='tgt_grease',
    version='1.2.5',
    description='GRE Application Service Engine',
    long_description="Automation Engine for operations",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Topic :: System',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7'
    ],
    keywords='python automated recovery incident',
    author='james.e.bell@target.com',
    packages=find_packages(),
    test_suite='nose.collector',
    tests_require=['nose', 'nose-cover3'],
    install_requires=[
        'psycopg2',
        'requests',
        'pymongo',
        'sqlalchemy',
        'python-dotenv'
    ] + (
         ["pypiwin32"] if "nt" == os.name else []
        ),
    include_package_data=True,
    zip_safe=False,
    scripts=[
        'bin/grease',
        'bin/grease-daemon',
        'bin/greasectl.py',
        'bin/greasectl.ps1'
    ]
)



from datetime import datetime, timedelta
from .BaseDetectorClass import BaseDetector
import operator


class TimeDelta(BaseDetector):
    def __init__(self):
        self._result = {'result': False}
        self.operators = {
            '>': operator.gt,
            'gt': operator.gt,
            '<': operator.lt,
            'lt': operator.lt,
            '>=': operator.ge,
            'ge': operator.ge,
            '<=': operator.le,
            'le': operator.le,
            '=': operator.eq,
            'eq': operator.eq,
            '!=': operator.ne,
            'ne': operator.ne
        }
        self.units = {
            'seconds': 'seconds',
            'second': 'seconds',
            'sec': 'seconds',
            'minutes': 'minutes',
            'minute': 'minutes',
            'min': 'minutes',
            'hours': 'hours',
            'hour': 'hours',
            'hr': 'hours',
            'days': 'days',
            'day': 'days',
            'years': 'years',
            'year': 'years',
            'yr': 'years'
        }

    def param_compute(self, source, rules):
        rules_passed = 0
        rules_needed_to_pass = len(rules)

        for delta_obj in rules:
            # Validation and setup
            # Mandatory checks
            if 'field' not in delta_obj or 'date_format' not in delta_obj or 'delta' not in delta_obj or \
               'unit' not in delta_obj or 'direction' not in delta_obj or 'operator' not in delta_obj:
                    self._result['result'] = False
                    return

            source_date = str(source[delta_obj['field']])
            date_format = str(delta_obj['date_format'])
            delta = abs(int(delta_obj['delta']))

            if str(delta_obj['unit']) in self.units:
                unit = self.units[str(delta_obj['unit'])]
            else:
                self._result['result'] = False
                return

            if str(delta_obj['direction']) is 'future':
                direction = 1
            else:
                direction = -1

            if str(delta_obj['operator']) in self.operators:
                op = self.operators[str(delta_obj['operator'])]
            else:
                self._result['result'] = False
                return

            if 'date' in delta_obj:
                try:
                    rel_dt = datetime.strptime(str(delta_obj['date']), date_format) + timedelta(**{unit: delta}) * direction
                except ValueError:
                    self._result['result'] = False
                    return
            else:
                rel_dt = datetime.now() + timedelta(**{unit: delta}) * direction

            try:
                source_dt = datetime.strptime(source_date, date_format)
            except ValueError:
                self._result['result'] = False
                return

            # Comparison
            # If direction is future, we need to have the source_dt on the left side
            # If it's past, we need source_dt on the right
            if str(delta_obj['direction']) is 'future':
                date_compare = op(source_dt, rel_dt)
            else:
                date_compare = op(rel_dt, source_dt)

            if date_compare:
                rules_passed += 1
                if 'variable' in delta_obj:
                    if delta_obj['variable'] and 'variable_name' in delta_obj:
                        self._result[delta_obj['variable_name']] = source_date
            else:
                # field not found in source
                # time to fail out since not all parameters can be met
                self._result = {'result': False}
                return

        if rules_passed >= rules_needed_to_pass:
            self._result['result'] = True
        else:
            self._result = {'result': False}

        return

    def get_result(self):
        return self._result



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, JSON, Boolean, ForeignKey, DateTime
from sqlalchemy.orm import relationship
from datetime import datetime


Base = declarative_base()


class JobConfig(Base):
    __tablename__ = "job_config"
    id = Column(Integer, primary_key=True, nullable=False)
    command_module = Column(String, nullable=False)
    command_name = Column(String, nullable=False)
    is_threaded = Column(Boolean, nullable=False, default=False)
    threads = Column(Integer, nullable=False, default=1)
    human_avg = Column(Integer, nullable=False, default=60)
    machine_avg = Column(Integer, nullable=False, default=1)
    tick = Column(Integer, nullable=False, default=1)


class JobServers(Base):
    __tablename__ = "job_servers"
    id = Column(Integer, primary_key=True, nullable=False)
    host_name = Column(String, nullable=False)
    execution_environment = Column(String, nullable=False, default="general")
    jobs_assigned = Column(Integer, nullable=False, default=0)
    detector = Column(Boolean, nullable=False, default=False)
    scheduler = Column(Boolean, nullable=False, default=False)
    active = Column(Boolean, nullable=False, default=True)
    activation_time = Column(DateTime, nullable=False, default=datetime.utcnow)


class JobQueue(Base):
    __tablename__ = "job_queue"
    id = Column(Integer, primary_key=True, nullable=False)
    host_name = Column(Integer, ForeignKey('job_servers.id'))
    ticket = Column(String)
    additional = Column(JSON, nullable=False)
    run_priority = Column(Integer, nullable=False, default=10)
    in_progress = Column(Boolean, nullable=False, default=False)
    completed = Column(Boolean, nullable=False, default=False)
    job_id = Column(Integer, ForeignKey('job_config.id'))
    request_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    complete_time = Column(DateTime)
    failures = Column(Integer, nullable=False, default=0)

    JobID = relationship(JobConfig, foreign_keys=[job_id])
    HostName = relationship(JobServers, foreign_keys=[host_name])

    def __repr__(self):
        return str(self.id)


class JobTelemetry(Base):
    __tablename__ = "job_telemetry"
    id = Column(Integer, primary_key=True, nullable=False)
    command = Column(Integer, ForeignKey('job_config.id'))
    affected = Column(Integer, nullable=False, default=0)
    start_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    entry_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    success = Column(Boolean, nullable=False, default=False)
    server_id = Column(Integer, ForeignKey('job_servers.id'))

    CommandID = relationship(JobConfig, foreign_keys=[command])
    ServerID = relationship(JobServers, foreign_keys=[server_id])


class JobTelemetryDaemon(Base):
    __tablename__ = "job_telemetry_daemon"
    id = Column(Integer, primary_key=True, nullable=False)
    command = Column(Integer, nullable=False)
    affected = Column(Integer, nullable=False, default=0)
    start_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    entry_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    execution_success = Column(Boolean, nullable=False, default=False)
    command_success = Column(Boolean, nullable=False, default=False)
    server_id = Column(Integer, ForeignKey('job_servers.id'))

    ServerID = relationship(JobServers, foreign_keys=[server_id])


class PersistentJobs(Base):
    __tablename__ = "persistent_jobs"
    id = Column(Integer, primary_key=True, nullable=False)
    server_id = Column(Integer, ForeignKey('job_servers.id'))
    command = Column(Integer, ForeignKey('job_config.id'))
    additional = Column(JSON, nullable=False)
    enabled = Column(Boolean, nullable=False, default=True)

    CommandID = relationship(JobConfig, foreign_keys=[command])
    ServerID = relationship(JobServers, foreign_keys=[server_id])


class SourceData(Base):
    __tablename__ = "source_data"
    id = Column(Integer, primary_key=True, nullable=False)
    source_data = Column(JSON, nullable=False)
    source_server = Column(Integer, ForeignKey('job_servers.id'), nullable=False)
    created_time = Column(DateTime, nullable=False)
    scanner = Column(String, nullable=False)
    detection_server = Column(Integer, ForeignKey('job_servers.id'), nullable=True)
    detection_start_time = Column(DateTime)
    detection_end_time = Column(DateTime)
    detection_complete = Column(Boolean, nullable=False, default=False)
    scheduling_server = Column(Integer, ForeignKey('job_servers.id'), nullable=True)
    scheduling_start_time = Column(DateTime)
    scheduling_end_time = Column(DateTime)
    scheduling_complete = Column(Boolean, nullable=False, default=False)

    SourceServer = relationship(JobServers, foreign_keys=[source_server])
    DetectionServer = relationship(JobServers, foreign_keys=[detection_server])
    SchedulingServer = relationship(JobServers, foreign_keys=[scheduling_server])


class ServerHealth(Base):
    __tablename__ = "server_health"
    id = Column(Integer, primary_key=True, nullable=False)
    server = Column(Integer, ForeignKey('job_servers.id'))
    job_hash = Column(String, nullable=False)
    check_time = Column(DateTime, nullable=False, default=datetime.utcnow)
    doctor = Column(Integer, ForeignKey('job_servers.id'), nullable=True)

    ServerNode = relationship(JobServers, foreign_keys=[server])
    DoctorNode = relationship(JobServers, foreign_keys=[doctor])


def __main__():
    print("INSTALLING DATABASE")
    from .Configuration import Configuration
    from .Database import SQLAlchemyConnection
    conf = Configuration()
    sql = SQLAlchemyConnection(conf)
    print(
        "INSTALLING DATABASE: [{0}] in location [{1}]".format(
            conf.get('GREASE_DB_ENGINE'),
            conf.get('GREASE_DB_HOST')
        )
    )
    Base.metadata.create_all(sql.get_engine())
    print("STARTING BASE JOB IMPORTS")
    Scanner = JobConfig(
        command_module='enterprise',
        command_name='engage_sensor_array',
        is_threaded=False,
        threads=1,
        human_avg=1,
        machine_avg=1,
        tick=0
    )
    Detectors = JobConfig(
        command_module='enterprise',
        command_name='engage_lcars',
        is_threaded=False,
        threads=1,
        human_avg=1,
        machine_avg=1,
        tick=21
    )
    Scheduler = JobConfig(
        command_module='enterprise',
        command_name='engage_warp',
        is_threaded=False,
        threads=1,
        human_avg=1,
        machine_avg=1,
        tick=41
    )
    ClusterHealth = JobConfig(
        command_module='enterprise',
        command_name='system_heartbeat',
        is_threaded=False,
        threads=1,
        human_avg=1,
        machine_avg=1,
        tick=50
    )
    print("ADDING SCANNER")
    sql.get_session().add(Scanner)
    sql.get_session().commit()
    print("ADDING DETECTOR")
    sql.get_session().add(Detectors)
    sql.get_session().commit()
    print("ADDING SCHEDULER")
    sql.get_session().commit()
    sql.get_session().add(Scheduler)
    sql.get_session().commit()
    print("ADDING CLUSTER MONITOR")
    sql.get_session().add(ClusterHealth)
    sql.get_session().commit()



import requests
import os
from urllib3.exceptions import HTTPError


class Notifier(object):

    def __init__(self):
        #Make sure we have the appropriate keys
        self._auth_token = os.getenv('GREASE_HIPCHAT_TOKEN', '')
        self._hipchat_room = os.getenv('GREASE_HIPCHAT_ROOM', '')
        self._base_url = 'https://api.hipchat.com/v2/room/'
        self._response = None

    def get_response(self):
        # type: () -> requests.Response
        return self._response

    def set_hipchat_room(self, room, token):
        #Each HipChat room has a unique auth token, so we have to change both.
        self._auth_token = token
        self._hipchat_room = room

    def send_hipchat_message(self, message, message_color='gray'):
        url = self._base_url + self._hipchat_room + '/notification?auth_token=' + self._auth_token 
        request_data = {
            'message': message,
            'color': message_color
        }

        try:
            self._response = requests.post(
                url=url,
                data=request_data,
                verify=False
            )
            if self.get_response().status_code == 204:
                return True
            else:
                return False
        except HTTPError as e:
            return False

